from ...exception_handling import handling_exceptions

class ActiveFixture(object):

    def __init__(self, fixture):
        super(ActiveFixture, self).__init__()
        self.fixture = fixture
        self.id = fixture.info.id
        self.name = fixture.info.name
        self._test_start_callbacks = []
        self._test_end_callbacks = []
        self._cleanups = []
        self._test_start_called = False

    def test_start(self, callback):
        self._test_start_callbacks.append(callback)

    def test_end(self, callback):
        self._test_end_callbacks.append(callback)

    def call_test_start(self):
        if self._test_start_called:
            return

        self._test_start_called = True
        with handling_exceptions():
            for callback in self._test_start_callbacks:
                callback()

    def call_test_end(self):
        self._test_start_called = False
        with handling_exceptions():
            for callback in self._test_end_callbacks:
                callback()

    def add_cleanup(self, cleanup):
        self._cleanups.append(cleanup)

    def do_cleanups(self):
        while self._cleanups:
            cleanup = self._cleanups.pop()
            cleanup()



# pylint: disable=unused-argument
import slash
import logbook

def test_retained_logs_handler(unique_string1):

    with slash.log.RetainedLogHandler() as handler:
        slash.logger.info(unique_string1)

    other_handler = logbook.TestHandler()
    handler.flush_to_handler(other_handler)

    [r] = other_handler.records # pylint: disable=unbalanced-tuple-unpacking
    assert unique_string1 in r.message


def test_disable():
    with slash.log.RetainedLogHandler() as handler:
        handler.disable()
        slash.logger.info('hey')
    assert not handler.records



class Details(object):

    def __init__(self, set_callback=None):
        super(Details, self).__init__()
        self._details = {}
        self._set_callback = set_callback

    def set(self, key, value):
        """Sets a specific detail (by name) to a specific value
        """
        if self._set_callback is not None:
            self._set_callback(key, value)
        self._details[key] = value

    def append(self, key, value):
        """Appends a value to a list key, or creates it if needed
        """
        lst = self._details.setdefault(key, [])
        if not isinstance(lst, list):
            raise TypeError('Cannot append value to a {.__class__.__name__!r} value'.format(lst))
        lst.append(value)
        if self._set_callback is not None:
            self._set_callback(key, lst)

    def all(self):
        return self._details.copy()

    def __nonzero__(self):
        return bool(self._details)

    __bool__ = __nonzero__

    def __contains__(self, key):
        return key in self._details



from __future__ import print_function

import threading
import time
import datetime
from contextlib import contextmanager
from .. import hooks
from ..conf import config
from ..ctx import context
from ..core import metadata
from ..core.function_test import FunctionTestFactory
from ..exceptions import TerminatedException
from ..exception_handling import handle_exception

from IPython.terminal.embed import embed  # pylint: disable=F0401

def _interact(ns):
    def _handle_exception(shell, exc_type, exc_value, exc_tb, tb_offset):
        exc_info = (exc_type, exc_value, exc_tb)
        shell.showtraceback(exc_info, tb_offset)
        if not _is_exception_in_ipython_eval(exc_tb):
            handle_exception(exc_info)
        if isinstance(exc_value, TerminatedException):
            context.result.add_error('Terminated')
            shell.exit_now = True

    embed(user_ns=ns, display_banner=False, custom_exceptions=((Exception, TerminatedException), _handle_exception))


def _is_exception_in_ipython_eval(exc_tb):
    while exc_tb.tb_next is not None:
        exc_tb = exc_tb.tb_next
    return exc_tb.tb_frame.f_code.co_filename.lower().startswith('<ipython')

def start_interactive_shell(**namespace):
    """
    Starts an interactive shell. Uses IPython if available, else fall back
    to the native Python interpreter.

    Any keyword argument specified will be available in the shell ``globals``.
    """
    if context.g is not None and config.root.interactive.expose_g_globals:
        namespace.update(context.g.__dict__)

    hooks.before_interactive_shell(namespace=namespace)  # pylint: disable=no-member
    _interact(namespace)

def _start_interactive_test():
    return start_interactive_shell()

def generate_interactive_test():
    [returned] = FunctionTestFactory(_start_interactive_test).generate_tests(context.session.fixture_store)
    returned.__slash__ = metadata.Metadata(None, returned)
    returned.__slash__.allocate_id()
    returned.__slash__.mark_interactive()
    returned.__slash__.set_file_path('<Interactive>')
    returned.__slash__.set_test_full_name('Interactive')
    returned.__slash__.factory_name = 'Interactive'
    return returned

def _humanize_time_delta(seconds):
    return str(datetime.timedelta(seconds=seconds)).partition('.')[0]

@contextmanager
def notify_if_slow_context(message, slow_seconds=1, end_message=None, show_duration=True):
    evt = threading.Event()
    evt.should_report_end_msg = False
    def notifier():
        if not evt.wait(timeout=slow_seconds) and context.session is not None:
            context.session.reporter.report_message(message)
            evt.should_report_end_msg = True
    thread = threading.Thread(target=notifier)
    start_time = time.time()
    thread.start()

    try:
        yield
    finally:
        evt.set()
        thread.join()
        if evt.should_report_end_msg and end_message is not None:
            if show_duration:
                end_message += ' (took {})'.format(_humanize_time_delta(time.time() - start_time))
            context.session.reporter.report_message(end_message)



# pylint: disable=redefined-outer-name
from .utils import TestCase
from .utils import run_tests_assert_success
import itertools
import os
import slash
from slash._compat import izip_longest
import pytest

from .utils.suite_writer import Suite


@pytest.mark.parametrize('parametrize', [True, False])
def test_class_name(suite, suite_test, test_type, parametrize):
    if parametrize:
        suite_test.add_parameter(num_values=3)
    summary = suite.run()
    for result in summary.get_all_results_for_test(suite_test):
        if test_type == 'method':
            assert result.test_metadata.class_name.startswith('Test')
            assert '(' not in result.test_metadata.class_name
        elif test_type == 'function':
            assert result.test_metadata.class_name is None
        else:
            raise NotImplementedError()  # pragma: no cover


@pytest.mark.parametrize('parametrize', [True, False])
def test_function_name(suite, suite_test, parametrize):
    if parametrize:
        suite_test.add_parameter(num_values=3)

    summary = suite.run()
    for result in summary.get_all_results_for_test(suite_test):
        function_name = result.test_metadata.function_name
        assert function_name.startswith('test_')
        assert '.' not in result.test_metadata.function_name
        assert '(' not in result.test_metadata.function_name


def test_variation(suite, suite_test):
    fixture = suite.slashconf.add_fixture()
    param = fixture.add_parameter()  # pylint: disable=unused-variable
    suite_test.depend_on_fixture(fixture)
    suite_test.append_line('slash.context.result.data["variation"] = slash.context.test.__slash__.variation.values.copy()')
    summary = suite.run()
    for result in summary.get_all_results_for_test(suite_test):
        assert len(result.data['variation']) == 1
        assert fixture.name not in result.data['variation']
        assert '{}.{}'.format(fixture.name, param.name) in result.data['variation']


def test_function_name_with_special_parameters(test_type):
    suite = Suite()
    assert len(suite) == 0  # pylint: disable=len-as-condition
    suite_test = suite.add_test(type=test_type)
    values = ['a.b', 'a(b']
    suite_test.add_parameter(values=values)

    # we can't verify result because we would not be able to parse the function properly
    # TODO: this will change once we properly support variations metadata  # pylint: disable=fixme
    summary = suite.run(verify=False, sort=False)
    for result, value in izip_longest(summary.session.results, values):
        function_name = result.test_metadata.function_name
        assert value not in function_name
        assert '.' not in result.test_metadata.function_name
        assert '(' not in result.test_metadata.function_name
        assert function_name.startswith('test_')


def test_module_name_not_none_or_empty_string(suite):
    for result in suite.run().session.results:
        assert result.test_metadata.module_name


def test_test_index(suite):
    index = None
    session = suite.run().session
    for index, result in enumerate(session.results):
        assert result.test_metadata.test_index0 == index
        assert result.test_metadata.test_index1 == index + 1
    assert index > 0


def test_set_test_name(test_metadata):
    assert test_metadata.file_path in str(test_metadata)
    custom_name = 'some_custom_name'
    test_metadata.set_test_full_name(custom_name)
    assert str(test_metadata) == '<{}>'.format(custom_name)


def test_class_name_with_dot_parameters():

    # pylint: disable=unused-argument

    @slash.parametrize('path', ['x.y'])
    def test_something(path):
        pass

    with slash.Session() as s:  # pylint: disable=unused-variable
        loader = slash.loader.Loader()
        [test] = loader.get_runnables(test_something)  # pylint: disable=unbalanced-tuple-unpacking
        assert test.__slash__.class_name is None


def test_set_file_path(test_metadata):
    file_path = '/tmp/file_path.py'
    assert file_path not in test_metadata.address
    test_metadata.set_file_path(file_path)
    assert test_metadata.file_path == file_path
    assert file_path in test_metadata.address


def test_mark_interactive(test_metadata):
    test_metadata.mark_interactive()
    assert test_metadata.is_interactive()


@pytest.fixture
def test_metadata(suite, suite_test):
    return suite.run()[suite_test].test_metadata


class TestMetadataTest(TestCase):
    loaded_tests = []

    def setUp(self):
        @slash.hooks.register
        def tests_loaded(tests): # pylint: disable=unused-variable
            TestMetadataTest.loaded_tests = tests

        super(TestMetadataTest, self).setUp()
        self.root = self.get_new_path()
        self.filename = os.path.join(self.root, "testfile.py")
        with open(self.filename, "w") as f:
            f.write(_TEST_FILE_TEMPLATE)

        with slash.Session() as s:
            self.session = run_tests_assert_success(self.filename, session=s)
            self.tests = self.loaded_tests
        self.results = list(self.session.results.iter_test_results())
        self.results.sort(key=lambda result: str(result.test_metadata))

    def test_tests_have_correct_metadata(self):
        for test, result in zip(self.tests, self.session.results.iter_test_results()):
            self.assertIs(test.__slash__, result.test_metadata)

    def test_simple_test_address(self):
        self.assertEqual(self.results[0].test_metadata.address, "{}:T001.test_method".format(self.filename))

    def test_parameterized_test_address(self):
        parameterized = set(x.test_metadata.address for x in self.results[1:])

        self.assertEqual(parameterized, set(
            "{0}:T002.test_parameters(after:c={2},b={3},before:a={1})".format(self.filename, a, c, b)
            for a, b, c in itertools.product([1, 2], [3, 4], [5, 6])))

_TEST_FILE_TEMPLATE = """
import slash

class T001(slash.Test):
    def test_method(self):
        pass

class T002(slash.Test):

    @slash.parameters.iterate(a=[1, 2])
    def before(self, a):
        pass

    @slash.parameters.iterate(b=[3, 4])
    def test_parameters(self, b):
        pass

    @slash.parameters.iterate(c=[5, 6])
    def after(self, c):
        pass
"""



import itertools

from .test import Test


class MethodTest(Test):  # pylint: disable=abstract-method

    def __init__(self, suite, cls):
        super(MethodTest, self).__init__(suite, cls.file)
        self.cls = cls

    def is_method_test(self):
        return True

    def _get_argument_strings(self):
        return itertools.chain(['self'], super(MethodTest, self)._get_argument_strings())



import itertools
import sys

from ..ctx import context

_sort_key_generator = itertools.count(1)

class Metadata(object):

    """Class representing the metadata associated with a test object. Generally available
    as test.__slash__
    """

    #: The index of the test in the current execution, 0-based
    test_index0 = None

    def __init__(self, factory, test):
        super(Metadata, self).__init__()
        #: The test's unique id
        self.id = None
        self.tags = test.get_tags()
        self._sort_key = next(_sort_key_generator)
        self.repeat_all_index = 0
        if factory is not None:
            #: The path to the file from which this test was loaded
            self.module_name = factory.get_module_name()
            assert self.module_name, 'Could not find module for {}'.format(test)
            self._file_path = factory.get_filename()
            self.factory_name = factory.get_factory_name()
        else:
            self.module_name = type(test).__module__
            self._file_path = sys.modules[self.module_name].__file__
            self.factory_name = '?'
        self.variation = test.get_variation()
        assert self.variation is not None

        self._address_override = None
        #: Address string to identify the test inside the file from which it was loaded
        self.address_in_file = self.factory_name
        self.address_in_factory = test.get_address_in_factory()
        if self.address_in_factory is not None:
            self.address_in_file += self.address_in_factory
        if factory is not None:
            self._class_name = factory.get_class_name()
        else:
            testfunc = test.get_test_function()
            if hasattr(testfunc, '__self__'):
                self._class_name = testfunc.__self__.__class__.__name__
            else:
                self._class_name = None

        self._interactive = False

    def set_file_path(self, file_path):
        self._file_path = file_path

    @property
    def file_path(self):
        return self._file_path

    def get_address(self, raw_params=False):
        """
        String identifying the test, to be used when logging or displaying
        results in the console generally it is composed of the file path and
        the address inside the file

        :param raw_params: If ``True``, emit the full parametrization values are interpolated into the returned
           string
        """
        if self._address_override is not None:
            return self._address_override

        returned = '{}:{}'.format(self.file_path, self.address_in_file)
        if self.variation:
            returned += '({})'.format(
                ', '.join('{}={!r}'.format(key, value) for key, value in self.variation.values.items())
                if raw_params
                else self.variation.safe_repr
            )
        return returned

    address = property(get_address)

    def allocate_id(self):
        assert self.id is None
        self.id = context.session.id_space.allocate()

    def set_sort_key(self, key):
        self._sort_key = key

    def get_sort_key(self):
        return self._sort_key

    def set_test_full_name(self, name):
        assert hasattr(self, '_address_override')
        self._address_override = name

    def is_interactive(self):
        return self._interactive

    def mark_interactive(self):
        self._interactive = True
        self.set_sort_key(0)

    @property
    def class_name(self):
        return self._class_name

    @property
    def function_name(self):
        returned = self.address_in_file
        if self._class_name:
            prefix = self._class_name + '.'
            assert returned.startswith(prefix)
            returned = returned[len(prefix):]

        return returned.split('(')[0]

    @property
    def test_index1(self):
        """Same as ``test_index0``, only 1-based
        """
        if self.test_index0 is None:
            return None
        return self.test_index0 + 1

    def __repr__(self):
        return '<{}>'.format(self.address)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.address == other.address
        return False

    def __ne__(self, other):
        return not (self == other)  # pylint: disable=superfluous-parens,unneeded-not

    def __hash__(self):
        return self.address.__hash__()

def ensure_test_metadata(test):
    returned = getattr(test, "__slash__", None)
    if returned is None:
        returned = test.__slash__ = Metadata(None, test)
    return returned



import pytest

@pytest.mark.parametrize('scope', ['module', 'test'])
def test_fixture_start_end_test(suite, suite_test, scope):

    fixture = suite.slashconf.add_fixture(scope=scope)

    start_event = fixture.add_deferred_event('this.test_start', name='fixture_test_start')
    end_event = fixture.add_deferred_event('this.test_end', name='fixture_test_end')
    test_event = suite_test.add_event()

    suite_test.depend_on_fixture(fixture)
    events = suite.run().events

    if scope != 'module':
        # for module scope, the event will get run over, resulting in later `test_start` events...
        assert events[start_event].timestamp < events[test_event].timestamp
    assert events[end_event].timestamp > events[test_event].timestamp



# pylint: disable=redefined-outer-name
import itertools
import os
import random
import tempfile
from uuid import uuid4

from forge import Forge

import gossip
import pytest
import slash
import slash.plugins
from slash.loader import Loader
from slash.core.result import GlobalResult, Result

from .utils.cartesian import Cartesian
from .utils.suite_writer import Suite
from .utils.suite_builder import SuiteBuilder
from .utils.garbage_collection import GarbageCollectionMarker


@pytest.fixture(scope='session', autouse=True)
def unittest_mode_logging():
    slash.config.root.log.unittest_mode = True


@pytest.fixture(scope='session', autouse=True)
def random_seed():
    random.seed(0xdeadface)


@pytest.fixture(scope='session', autouse=True)
def no_user_config(request):
    tmpdir = tempfile.mkdtemp()
    slash.conf.config.root.run.user_customization_file_path = os.path.join(
        tmpdir, 'slashrc')

    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        os.rmdir(tmpdir)


@pytest.fixture
def no_plugins(request):
    slash.plugins.manager.uninstall_all()
    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        slash.plugins.manager.uninstall_all()
        slash.plugins.manager.install_builtin_plugins()


@pytest.fixture
def forge(request):

    returned = Forge()

    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        returned.verify()
        returned.restore_all_replacements()

    return returned


@pytest.fixture
def config_override(request):

    def _override(path, value):
        prev_value = slash.config.get_config(path).get_value()

        @request.addfinalizer
        def restore():  # pylint: disable=unused-variable
            slash.config.assign_path(path, prev_value)
        slash.config.assign_path(path, value)
    return _override


@pytest.fixture
def cartesian():
    return Cartesian()


@pytest.fixture(scope="function", autouse=True)
def cleanup_hook_registrations(request):
    @request.addfinalizer
    def _cleanup():
        for hook in gossip.get_group("slash").get_hooks():
            hook.unregister_all()
        assert not gossip.get_group("slash").get_subgroups()


@pytest.fixture
def checkpoint():
    return Checkpoint()

@pytest.fixture
def checkpoint1():
    return Checkpoint()

@pytest.fixture
def checkpoint2():
    return Checkpoint()


_timestamp = itertools.count(1000000)


class Checkpoint(object):

    called_count = 0
    args = kwargs = timestamp = None

    def __call__(self, *args, **kwargs):
        self.called_count += 1
        self.args = args
        self.kwargs = kwargs
        self.timestamp = next(_timestamp)

    @property
    def called(self):
        return self.called_count > 0

@pytest.fixture
def suite_test(suite, test_type, is_last_test):
    returned = suite.add_test(type=test_type)
    if not is_last_test:
        _ = suite.add_test(type=test_type)

    return returned

@pytest.fixture
def last_suite_test(suite, test_type):
    return suite.add_test(type=test_type)

@pytest.fixture(params=[GlobalResult, Result])
def result(request):
    return request.param()


@pytest.fixture(params=[True, False])
def is_last_test(request):
    return request.param


@pytest.fixture(params=['method', 'function'])
def test_type(request):
    return request.param


@pytest.fixture
def suite():
    returned = Suite()
    returned.populate()
    return returned


@pytest.fixture
def suite_builder(tmpdir):
    return SuiteBuilder(str(tmpdir.join('suite_builder')))


@pytest.fixture
def parallel_suite_test(parallel_suite, test_type, is_last_test):
    returned = parallel_suite.add_test(type=test_type)
    if not is_last_test:
        _ = parallel_suite.add_test(type=test_type)

    return returned


@pytest.fixture
def parallel_suite():
    returned = Suite(debug_info=False, is_parallel=True)
    returned.populate()
    return returned

@pytest.fixture
def runnable_test_dir(tmpdir):
    tests_dir = tmpdir.join(str(uuid4()))
    filename = str(uuid4()).replace('-', '') + '.py'
    with tests_dir.join(filename).open('w', ensure=True) as f:
        f.write('def test_something():\n    pass')
    return tests_dir

@pytest.fixture
def slash_session():
    return slash.Session()

@pytest.fixture
def test_loader():
    return Loader()


@pytest.fixture
def active_slash_session(request):
    returned = slash.Session()
    returned.__enter__()

    @request.addfinalizer
    def finalize():  # pylint: disable=unused-variable
        returned.__exit__(None, None, None)

    return returned


@pytest.fixture(params=["slashconf", "module"])
def defined_fixture(request, suite, suite_test):
    if request.param == 'slashconf':
        return suite.slashconf.add_fixture()
    elif request.param == 'module':
        return suite_test.file.add_fixture()

    raise NotImplementedError()  # pragma: no cover


@pytest.fixture
def gc_marker():
    return GarbageCollectionMarker()


@pytest.fixture(autouse=True, scope="function")
def reset_gossip(request):
    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        for group in list(gossip.get_groups()):
            if group.name == 'slash':
                continue
            group.undefine()

        for hook in gossip.get_all_hooks():
            if hook.group.name != 'slash':
                hook.undefine()
            else:
                hook.unregister_all()


@pytest.fixture  # pylint: disable=unused-argument
def plugin(no_plugins):  # pylint: disable=unused-argument

    class StartSessionPlugin(slash.plugins.PluginInterface):
        _activate_called = False
        _deactivate_called = False

        def __init__(self):
            super(StartSessionPlugin, self).__init__()
            self.session_start_call_count = 0

        def get_name(self):
            return "start-session"

        def session_start(self):
            self.session_start_call_count += 1

        def activate(self):
            self._activate_called = True

        def deactivate(self):
            self._deactivate_called = True
    return StartSessionPlugin()


@pytest.fixture(params=['slashconf', 'file'])
def get_fixture_location(request):

    def getter(test):
        if request.param == 'slashconf':
            return test.suite.slashconf
        elif request.param == 'file':
            return test.file
        else:
            raise NotImplementedError() # pragma: no cover
    return getter

@pytest.fixture
def restore_plugins_on_cleanup(request):
    request.addfinalizer(slash.plugins.manager.install_builtin_plugins)
    request.addfinalizer(slash.plugins.manager.uninstall_all)


@pytest.fixture
def logs_dir(config_override, tmpdir, relative_symlinks):
    returned = tmpdir.join('logs')
    config_override("log.root", str(returned.join("files")))
    config_override("log.last_session_symlink",
                    str("../links/last-session" if relative_symlinks else returned.join("links", "last-session")))
    config_override("log.last_session_dir_symlink",
                    str("../links/last-session-dir" if relative_symlinks else returned.join("links", "last-session-dir")))
    config_override("log.last_test_symlink",
                    str("../links/last-test" if relative_symlinks else returned.join("links", "last-test")))
    config_override("log.last_failed_symlink",
                    str("../links/last-failed" if relative_symlinks else returned.join("links", "last-failed")))

    return returned


@pytest.fixture(params=[True, False])
def relative_symlinks(request):
    return request.param


@pytest.fixture
def session_log(logs_dir):
    return logs_dir.join('links').join('last-session')


@pytest.fixture
def unique_string1():
    return str(uuid4())


@pytest.fixture(params=[True, False])
def yield_fixture_decorator(request):
    should_use_explicitly = request.param
    if should_use_explicitly:
        return slash.yield_fixture
    return slash.fixture


@pytest.fixture  # pylint: disable=unused-argument
def xunit_filename(tmpdir, request, config_override):  # pylint: disable=unused-argument
    xunit_filename = str(tmpdir.join('xunit.xml'))
    slash.plugins.manager.activate('xunit')

    slash.config.root.plugin_config.xunit.filename = xunit_filename

    @request.addfinalizer
    def deactivate():  # pylint: disable=unused-variable
        slash.plugins.manager.deactivate('xunit')

    return xunit_filename



import logbook
import signal
import sys
import dessert
from contextlib import contextmanager

from . import hooks as trigger_hook
from . import log
from . import site
from . import plugins
from ._compat import ExitStack
from .conf import config
from .core.session import Session
from .reporting.console_reporter import ConsoleReporter
from . import exceptions
from .exceptions import TerminatedException, SlashException
from .exception_handling import handling_exceptions, inhibit_unhandled_exception_traceback, should_inhibit_unhandled_exception_traceback
from .loader import Loader
from .log import ConsoleHandler
from .utils import cli_utils
from .utils.debug import debug_if_needed

_logger = logbook.Logger(__name__)
inhibit_unhandled_exception_traceback(SlashException)


class Application(object):

    def __init__(self):
        super(Application, self).__init__()
        self._test_loader = Loader()
        self.set_report_stream(sys.stderr)
        self._argv = None
        self._reset_parser()
        self._positional_args = None
        self._parsed_args = None
        self._reporter = None
        self.test_loader = Loader()
        self.session = None
        self._working_directory = None
        self._interrupted = False
        self._exit_code = 0

    def set_working_directory(self, path):
        self._working_directory = path

    @property
    def exit_code(self):
        return self._exit_code

    def set_exit_code(self, exit_code):
        self._exit_code = exit_code

    @property
    def interrupted(self):
        return self._interrupted

    @property
    def positional_args(self):
        return self._positional_args

    @property
    def parsed_args(self):
        return self._parsed_args

    def enable_interactive(self):
        self.arg_parser.add_argument(
            '-i', '--interactive', help='Enter an interactive shell',
            action="store_true", default=False)

    def _reset_parser(self):
        self.arg_parser = cli_utils.SlashArgumentParser()

    def set_argv(self, argv):
        self._argv = list(argv)

    def _get_argv(self):
        if self._argv is None:
            return sys.argv[1:]
        return self._argv[:]

    def set_report_stream(self, stream):
        if stream is not None:
            self._report_stream = stream
            self._default_reporter = ConsoleReporter(level=logbook.ERROR, stream=stream)
            self._console_handler = ConsoleHandler(stream=stream, level=logbook.ERROR)

    def set_reporter(self, reporter):
        self._reporter = reporter

    def get_reporter(self):
        returned = self._reporter
        if returned is None:
            returned = ConsoleReporter(
                level=config.root.log.console_level,
                stream=self._report_stream)

        return returned

    def __enter__(self):
        self._exit_stack = ExitStack()
        self._exit_stack.__enter__()
        try:
            self._exit_stack.enter_context(self._prelude_logging_context())
            self._exit_stack.enter_context(self._sigterm_context())
            with dessert.rewrite_assertions_context():
                site.load(working_directory=self._working_directory)

            cli_utils.configure_arg_parser_by_plugins(self.arg_parser)
            cli_utils.configure_arg_parser_by_config(self.arg_parser)
            argv = cli_utils.add_pending_plugins_from_commandline(self._get_argv())

            self._parsed_args, self._positional_args = self.arg_parser.parse_known_args(argv)

            self._exit_stack.enter_context(
                cli_utils.get_modified_configuration_from_args_context(self.arg_parser, self._parsed_args)
                )

            self.session = Session(reporter=self.get_reporter(), console_stream=self._report_stream)

            trigger_hook.configure() # pylint: disable=no-member
            plugins.manager.configure_for_parallel_mode()
            plugins.manager.activate_pending_plugins()
            cli_utils.configure_plugins_from_args(self._parsed_args)


            self._exit_stack.enter_context(self.session)
            self._emit_prelude_logs()
            return self

        except:
            self._emit_prelude_logs()
            self.__exit__(*sys.exc_info())
            raise

    def __exit__(self, exc_type, exc_value, exc_tb):
        exc_info = (exc_type, exc_value, exc_tb)
        debug_if_needed(exc_info)
        if exc_value is not None:
            self._exit_code = exc_value.code if isinstance(exc_value, SystemExit) else -1

            if should_inhibit_unhandled_exception_traceback(exc_value):
                self.get_reporter().report_error_message(str(exc_value))

            elif isinstance(exc_value, Exception):
                _logger.error('Unexpected error occurred', exc_info=exc_info)
                self.get_reporter().report_error_message('Unexpected error: {}'.format(exc_value))

            if isinstance(exc_value, exceptions.INTERRUPTION_EXCEPTIONS):
                self._interrupted = True

        if exc_type is not None:
            trigger_hook.result_summary() # pylint: disable=no-member
        self._exit_stack.__exit__(exc_type, exc_value, exc_tb)
        self._exit_stack = None
        self._reset_parser()
        return True

    def _prelude_logging_context(self):
        self._prelude_log_handler = log.RetainedLogHandler(bubble=True, level=logbook.TRACE)
        return self._prelude_log_handler.applicationbound()

    def _emit_prelude_logs(self):
        self._prelude_log_handler.disable()
        handler = None
        if self.session is not None:
            handler = self.session.logging.session_log_handler
        if handler is None:
            handler = self._console_handler
        self._prelude_log_handler.flush_to_handler(handler)

    @contextmanager
    def _sigterm_context(self):
        def handle_sigterm(*_):
            with handling_exceptions():
                raise TerminatedException('Terminated by signal')

        prev = signal.signal(signal.SIGTERM, handle_sigterm)
        try:
            yield
        finally:
            try:
                signal.signal(signal.SIGTERM, prev)
            except TypeError as e:
                #workaround for a strange issue on app cleanup. See https://bugs.python.org/issue23548
                if 'signal handler must be signal.SIG_IGN' not in str(e):
                    raise



# pylint: disable=unused-argument,redefined-outer-name
import gc

import pytest
from slash.utils.python import get_arguments


def test_gc_marker(gc_marker):

    class Obj(object):
        pass

    obj = Obj()
    marker = gc_marker.mark(obj)
    assert not marker.destroyed
    del obj
    gc.collect()
    assert marker.destroyed


def test_get_arguments(func):
    args = get_arguments(func)
    assert [a.name for a in args] == ['a', 'b', 'c']



class SampleClass(object):

    def example_func(self, a, b, c):
        pass

def example_func(a, b, c):
    pass

@pytest.fixture(params=[SampleClass.example_func, SampleClass().example_func, example_func])
def func(request):
    return request.param



from numbers import Number
import string

from .._compat import string_types

_PRINTABLE_TYPES = (Number,) + string_types
_PRINTABLE_CHARS = set(string.ascii_letters) | set(string.digits) | set("-_")


class Variation(object):

    """Represents a single variation of parameter points. A variation is merely a mapping of fixture ids to their values.
    This mostly applies for parametrization fixtures. The other fixtures follow since they are either constant
    or indirectly depend on parametrization"""

    def __init__(self, store, param_value_indices, param_name_bindings):
        """
        :param name_bindings: dictionary mapping parameter name to its id
        """
        super(Variation, self).__init__()
        self._store = store
        self.param_value_indices = param_value_indices
        self.id = {}
        self.values = {}
        self.labels = {}
        for param_name, param in param_name_bindings.items():
            value_index = self.id[param_name] = param_value_indices[param.info.id]
            self.values[param_name] = param.get_value_by_index(value_index)
            self.labels[param_name] = param.values[value_index].label
        self.safe_repr = self._get_safe_repr()

    def _get_safe_repr(self):
        returned = {}
        for name, value in self.values.items():
            returned[name] = self._format_parameter_value_safe(name, value)
        return ','.join('{}={}'.format(key, returned[key]) for key in sorted(returned))

    def _format_parameter_value_safe(self, name, value):
        label = self.labels[name]

        if isinstance(label, str):
            return label
        value = str(value)
        if self._is_printable(value):
            return str(value)
        return '{}{}'.format(name, label)

    def _is_printable(self, value):
        if not isinstance(value, _PRINTABLE_TYPES):
            return False
        return _PRINTABLE_CHARS.issuperset(value)

    def has_value_for_parameter(self, param_id):
        return param_id in self.param_value_indices

    def get_param_value(self, param):
        return self._store.get_value(self, param)

    def __eq__(self, other):
        if isinstance(other, Variation):
            other = other.param_value_indices
        if not isinstance(other, dict):
            return NotImplemented
        return self.param_value_indices == other

    def __ne__(self, other):
        return not (self == other)  # pylint: disable=superfluous-parens,unneeded-not

    def __repr__(self):
        return 'Variation({})'.format(', '.join('{}={}'.format(key, value) for key, value in self.param_value_indices.items()))

    def __nonzero__(self):
        return bool(self.param_value_indices)

    __bool__ = __nonzero__



# pylint: disable=unused-argument
import os

import gossip
import pytest
import slash
from slash._compat import PY2
from slash import hooks, plugins
from slash.plugins import IncompatiblePlugin, PluginInterface

from .utils import NamedPlugin, maybe_decorate



def test_registers_on_none(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active  # pylint: disable=unused-variable
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on(None)
        def some_method_here(self):
            checkpoint()

    gossip.trigger('slash.some_method_here')
    assert not checkpoint.called


@pytest.mark.parametrize('class_level_needs', [True, False])
@pytest.mark.parametrize('class_level_provides', [True, False])
def test_registers_on_kwargs(class_level_needs, class_level_provides):

    needs_decorator = plugins.needs('other_requirement')
    provides_decorator = plugins.provides('another_provided_requirement')

    @slash.plugins.active  # pylint: disable=unused-variable
    @maybe_decorate(needs_decorator, class_level_needs)
    @maybe_decorate(provides_decorator, class_level_provides)
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on('some.hook', provides=['provided_requirement'], needs=['some_requirement'], tags=['tag'])
        @maybe_decorate(needs_decorator, not class_level_needs)
        @maybe_decorate(provides_decorator, not class_level_provides)
        def plugin_method(self):
            pass


    @gossip.register('some.hook', provides=['some_requirement', 'other_requirement'])
    def _unused():
        pass

    gossip.trigger('some.hook')
    hook = gossip.get_hook('some.hook')
    [registration] = [reg for reg in hook.get_registrations() if reg.func.__name__ == 'plugin_method']
    assert registration.tags == {'tag'}
    assert registration.needs == frozenset(['some_requirement', 'other_requirement'])
    assert registration.provides == frozenset(['provided_requirement', 'another_provided_requirement'])


def test_registers_on_with_private_methods(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active  # pylint: disable=unused-variable
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on('some_hook')
        def _handler(self):
            checkpoint()

    assert not checkpoint.called
    gossip.trigger('some_hook')
    assert checkpoint.called


def test_class_variables_allowed(restore_plugins_on_cleanup):
    @slash.plugins.active  # pylint: disable=unused-variable
    class SamplePlugin(PluginInterface):

        ATTRIBUTE = 'some_value'

        def get_name(self):
            return 'sample'


def test_active_decorator(restore_plugins_on_cleanup):

    plugins.manager.uninstall_all()

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

    assert isinstance(SamplePlugin, type)
    assert issubclass(SamplePlugin, PluginInterface)
    [active] = plugins.manager.get_active_plugins().values()
    assert isinstance(active, SamplePlugin)


@pytest.mark.parametrize('is_internal', [True, False])
def test_custom_hook_registration(request, is_internal):

    hook_name = 'some_hook'
    with pytest.raises(LookupError):
        gossip.get_hook(hook_name)

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass
    p = MyPlugin()
    plugins.manager.install(p, activate=True, is_internal=is_internal)

    @request.addfinalizer
    def cleanup():              # pylint: disable=unused-variable
        plugins.manager.uninstall(p)

    registrations = gossip.get_hook(hook_name).get_registrations()
    assert len(registrations) == 1
    [r] = registrations
    if PY2:
        assert r.func.__func__ is MyPlugin.unknown.__func__  # pylint: disable=no-member
    else:
        assert r.func.__func__ is MyPlugin.unknown

    # make sure we deactivate properly as well
    plugins.manager.deactivate(p)
    assert not gossip.get_hook(hook_name).get_registrations()


def test_multiple_registers_on(request):
    hook_names = ['some_hook_{}'.format(i) for i in range(2)]

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_names[0])
        @plugins.registers_on(hook_names[1])
        def unknown(self):
            pass

    expected_func = MyPlugin.unknown.__func__ if PY2 else MyPlugin.unknown
    p = MyPlugin()
    plugins.manager.install(p, activate=True)
    @request.addfinalizer
    def cleanup():              # pylint: disable=unused-variable
        plugins.manager.uninstall(p)

    for hook_name in hook_names:
        registrations = gossip.get_hook(hook_name).get_registrations()
        assert len(registrations) == 1
        assert registrations[0].func.__func__ is expected_func

    plugins.manager.deactivate(p)

    for hook_name in hook_names:
        assert not gossip.get_hook(hook_name).get_registrations()

def test_register_invalid_hook():

    initially_installed = list(plugins.manager.get_installed_plugins())

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert set(plugins.manager.get_installed_plugins()) == set(initially_installed)


def test_register_custom_hooks_strict_group():

    initially_installed = list(plugins.manager.get_installed_plugins())

    hook_name = "some_group.some_hook"
    gossip.get_or_create_group("some_group").set_strict()

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert list(plugins.manager.get_installed_plugins()) == initially_installed


def test_builtin_plugins_hooks_start_condition():
    "make sure that all hooks are either empty, or contain callbacks marked with `slash.<identifier>`"
    for hook_name, hook in hooks.get_all_hooks():  # pylint: disable=unused-variable
        for registration in hook.get_registrations():
            assert registration.token.startswith('slash.'), 'Callback {} is not a builtin!'.format(hook.full_name)

def test_builtin_plugins_are_installed():
    installed = plugins.manager.get_installed_plugins()
    assert installed
    for filename in os.listdir(os.path.join(os.path.dirname(plugins.__file__), "builtin")):
        if filename.startswith("_") or filename.startswith(".") or not filename.endswith(".py"):
            continue
        assert filename[:(-3)] in installed


def test_get_installed_plugins():

    class CustomPlugin(PluginInterface):
        def __init__(self, name):
            super(CustomPlugin, self).__init__()
            self._name = name

        def get_name(self):
            return self._name

    some_plugin = CustomPlugin('some-plugin')
    internal_plugin = CustomPlugin('internal-plugin')
    plugins.manager.install(some_plugin)
    plugins.manager.install(internal_plugin, is_internal=True)

    assert some_plugin.get_name() in plugins.manager.get_installed_plugins(include_internals=True)
    assert some_plugin.get_name() in plugins.manager.get_installed_plugins(include_internals=False)
    assert internal_plugin.get_name() in plugins.manager.get_installed_plugins(include_internals=True)
    assert internal_plugin.get_name() not in plugins.manager.get_installed_plugins(include_internals=False)

def test_cannot_install_incompatible_subclasses(no_plugins):

    class Incompatible(object):
        pass
    for invalid in (Incompatible, Incompatible(), PluginInterface, object(), 1, "string"):
        with pytest.raises(IncompatiblePlugin):
            plugins.manager.install(invalid)
    assert plugins.manager.get_installed_plugins() == {}


def test_install_uninstall(no_plugins):
    plugin_name = "some_plugin_name"

    class CustomPlugin(PluginInterface):

        def get_name(self):
            return plugin_name
    with pytest.raises(LookupError):
        plugins.manager.get_plugin(plugin_name)
    plugin = CustomPlugin()
    plugins.manager.install(plugin)
    assert plugins.manager.get_plugin(plugin_name) is plugin
    plugins.manager.uninstall(plugin)
    with pytest.raises(LookupError):
        plugins.manager.get_plugin(plugin_name)



@pytest.mark.parametrize('cond', [True, False])
def test_register_if(no_plugins, checkpoint, cond):

    @slash.plugins.active  # pylint: disable=unused-variable
    class CustomPlugin(NamedPlugin):

        @slash.plugins.register_if(cond)
        def test_start(self):
            checkpoint()

    slash.hooks.test_start()  # pylint: disable=no-member

    assert checkpoint.called == cond


def test_register_if_nonexistent_hook(no_plugins, checkpoint):

    @slash.plugins.active  # pylint: disable=unused-variable
    class CustomPlugin(NamedPlugin):

        @slash.plugins.register_if(False)
        def nonexistent_hook(self):
            checkpoint()


def test_restoring_state_context():

    class Plugin1(NamedPlugin):
        pass

    @slash.plugins.active
    class Plugin2(NamedPlugin):
        pass

    class Plugin3(NamedPlugin):
        pass

    @slash.plugins.active
    class Plugin4(NamedPlugin):
        pass

    class Plugin5(NamedPlugin):
        pass


    manager = slash.plugins.manager
    manager.install(Plugin3())
    installed = manager.get_installed_plugins().copy()
    active = manager.get_active_plugins().copy()


    with manager.restoring_state_context():
        manager.install(Plugin1())
        manager.uninstall(Plugin2)
        manager.activate('Plugin3')
        manager.deactivate(Plugin4)
        manager.install(Plugin5(), activate=True)
        assert set(manager.get_installed_plugins()) == set(installed)\
            .union({'Plugin1', 'Plugin5'})\
            .difference({'Plugin2'})
        assert set(manager.get_active_plugins()) == set(active).union({'Plugin3', 'Plugin5'}).difference({'Plugin2', 'Plugin4'})
    assert manager.get_installed_plugins() == installed
    assert manager.get_active_plugins() == active



# pylint: disable=redefined-outer-name
import sys

import pytest

import slash
from slash.exceptions import INTERRUPTION_EXCEPTIONS
from slash.utils import debug


def test_debug_if_needed_regular_exception(replaced_checkpoint, exc_info, debug_enabled):  # pylint: disable=unused-argument
    with slash.Session():
        debug.debug_if_needed(exc_info)
    assert replaced_checkpoint.called

def test_debug_if_needed_not_called(replaced_checkpoint, skipped_exc_info, debug_enabled):  # pylint: disable=unused-argument
    with slash.Session():
        debug.debug_if_needed(skipped_exc_info)
    assert not replaced_checkpoint.called


@pytest.mark.parametrize('filter_strings,should_pdb', [
    (['division by zero',], True),
    (["not 'division by zero'",], False),
    (['ZeroDivisionError',], True),
    (['not ZeroDivisionError',], False),
    (['Zero', 'by'], True),
    (['Zero', 'foobar'], False),
    (['ZERO',], False),
])
def test_pdb_filtering(filter_strings, should_pdb, replaced_checkpoint, exc_info, config_override):
    config_override('debug.enabled', True)
    config_override('debug.filter_strings', filter_strings)
    with slash.Session():
        debug.debug_if_needed(exc_info)
    assert replaced_checkpoint.called == should_pdb


def test_pdb_filtering_with_disabled_debug(replaced_checkpoint, exc_info, config_override):
    config_override('debug.enabled', False)
    config_override('debug.filter_strings', ['ZeroDivisionError'])
    with slash.Session():
        debug.debug_if_needed(exc_info)
    assert not replaced_checkpoint.called


@pytest.fixture
def debug_enabled(config_override):
    config_override('debug.enabled', True)


@pytest.fixture(params=(SystemExit,) + INTERRUPTION_EXCEPTIONS)
def skipped_exc_info(request):
    try:
        raise request.param()
    except:
        return sys.exc_info()

@pytest.fixture
def exc_info():
    try:
        1/0
    except:
        return sys.exc_info()

@pytest.fixture
def replaced_checkpoint(checkpoint, forge):
    forge.replace_with(debug, 'launch_debugger', checkpoint)
    return checkpoint



import sys
import pytest

from slash.core.error import Error
from .utils import Unprintable


def test_exception_attributes(error):  # pylint: disable=redefined-outer-name, unused-argument
    attrs = error.exception_attributes
    assert attrs['x'] == 2
    assert isinstance(attrs['y'], str)


def test_exception_attributes_with_message():
    assert Error('x').exception_attributes is None


@pytest.fixture
def error():

    class MyException(Exception):

        def __init__(self):
            super(MyException, self).__init__()
            self.x = 2
            self.y = [str(i) for i in range(10)]
            self.d = {'some': 'dict', 'here': 'too'}
            self.unprintable = Unprintable()

    def func():
        raise MyException()

    try:
        func()
    except MyException:
        return Error(exc_info=sys.exc_info())
    else:
        assert False, 'Did not raise'



"""Slash should support running ad-hoc test objects, that were not loaded from a file
"""
import pytest
import slash


@pytest.mark.parametrize('is_class', [True, False])
def test_ad_hoc_test_running(checkpoint, is_class):

    if is_class:
        class test_sample(slash.Test):

            def test_something(self):
                checkpoint()

    else:
        def test_sample():
            checkpoint()

    with slash.Session() as session:
        with session.get_started_context():
            slash.runner.run_tests(slash.loader.Loader().get_runnables(test_sample))
    assert session.results.is_success()
    assert checkpoint.called



import collections
import itertools

from orderedset import OrderedSet

from .variation import Variation
from .._compat import OrderedDict, izip, xrange
from ..exceptions import FixtureException
from ..utils.python import get_arguments
from .fixtures.parameters import iter_parametrization_fixtures
from .fixtures.fixture import Fixture
from .fixtures.parameters import Parametrization
from .fixtures.utils import nofixtures


class VariationFactory(object):

    """Helper class to produce variations, while properly naming the needed fixtures to help identifying tests
    """

    def __init__(self, fixture_store):
        super(VariationFactory, self).__init__()
        self._store = fixture_store
        self._autouse_fixtures = list(fixture_store.iter_autouse_fixtures_in_namespace())
        self._needed_fixtures = list(self._autouse_fixtures)

        self._arg_name_bindings = OrderedDict()
        self._param_name_bindings = OrderedDict()
        self._known_value_strings = collections.defaultdict(dict)

    def add_needed_fixture_id(self, fixture_id):
        self._needed_fixtures.append(self._store.get_fixture_by_id(fixture_id))

    def add_needed_fixtures_from_method(self, method):
        self._add_needed_fixtures_from_function(method)

    def add_needed_fixtures_from_function(self, func):
        self._add_needed_fixtures_from_function(func)

    def _add_needed_fixtures_from_function(self, func):

        if isinstance(func, tuple):
            namespace, func = func
        else:
            namespace = None

        if nofixtures.is_marked(func):
            return

        args = get_arguments(func)

        parametrizations = {}
        for name, param in iter_parametrization_fixtures(func):
            # make sure the parametrization is in the store
            self._store.ensure_known_parametrization(param)
            parametrizations[name] = param

            self._needed_fixtures.append(param)

        for argument in args:
            fixture = parametrizations.get(argument.name, None)
            if fixture is None:
                try:
                    fixture = self._store.get_fixture_by_argument(argument)
                except FixtureException as e:
                    raise type(e)('Loading {0.__code__.co_filename}:{0.__name__}: {1}'.format(func, e))


            self._needed_fixtures.append(fixture)

            arg_name = argument.name
            if namespace is not None:
                arg_name = '{}:{}'.format(namespace, arg_name)

            self._populate_param_name_bindings(arg_name, fixture)
            self._arg_name_bindings[arg_name] = fixture

        for fixture in self._autouse_fixtures:
            self._populate_param_name_bindings(fixture.info.name, fixture, prefix='::')

    def _populate_param_name_bindings(self, arg_name, fixture_or_param, prefix=''):
        visited = {fixture_or_param.info.id}
        stack = [(prefix + arg_name, fixture_or_param)]
        while stack:
            name, fixture = stack.pop()
            if isinstance(fixture, Fixture):
                for sub_name, obj in fixture.keyword_arguments.items():
                    if obj.info.id in visited:
                        continue
                    visited.add(obj.info.id)
                    stack.append(('{}.{}'.format(name, sub_name), obj))
            elif isinstance(fixture, Parametrization):
                self._param_name_bindings[name] = fixture
            else:
                raise NotImplementedError() # pragma: no cover

    def iter_variations(self):
        needed_ids = OrderedSet()

        for fixture in self._needed_fixtures:
            needed_ids.update(self._store.get_all_needed_fixture_ids(fixture))

        parametrizations = [self._store.get_fixture_by_id(param_id) for param_id in needed_ids]
        if not needed_ids:
            yield Variation(self._store, {}, {})
            return
        for value_indices in itertools.product(*(xrange(len(p.values)) for p in parametrizations)):
            yield self._build_variation(parametrizations, value_indices)

    def _build_variation(self, parametrizations, value_indices):
        value_index_by_id = {}
        for param, param_index in izip(parametrizations, value_indices):
            value_index_by_id[param.info.id] = param_index

        return Variation(self._store, value_index_by_id, self._param_name_bindings)



import warnings

import pytest

import slash
from slash import Session

from slash.utils.warning_capture import warning_callback_context

def test_warning_capture_context():
    captured = []

    def callback(*args, **kwargs):
        captured.append(Warning(*args, **kwargs))

    with warning_callback_context(callback):
        warnings.warn('some warning')

    assert len(captured) == 1
    [w] = captured # pylint: disable=unbalanced-tuple-unpacking
    assert w.args[0].args[0] == 'some warning'


def test_session_adds_simple_filter(request):
    @request.addfinalizer
    def cleanup():              # pylint: disable=unused-variable
        warnings.simplefilter('default')
    warnings.simplefilter('ignore')
    with Session() as s:
        warnings.warn('bla')

    assert len(s.warnings.warnings) == 1

def test_session_warning_calls_hook():
    notified = []

    @slash.hooks.warning_added.register # pylint: disable=no-member
    def warning_added(warning):         # pylint: disable=unused-variable
        notified.append(warning)

    with Session() as session:
        warnings.warn('bla')

    assert list(session.warnings) == notified != []


def test_cant_reenter_twice():
    ctx = warning_callback_context(no_op)
    with ctx:
        with pytest.raises(RuntimeError):
            with ctx:
                pass

def test_cant_exit_no_enter():
    ctx = warning_callback_context(no_op)
    with pytest.raises(RuntimeError):
        ctx.__exit__(None, None, None)


def no_op(*args, **kwargs):     # pylint: disable=unused-argument
    pass

class Warning(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs



from contextlib import contextmanager
import slash
import py.code  # pylint: disable=no-name-in-module, import-error


def get_code_lines(function):
    source_lines = str(py.code.Code(function).source()).splitlines()  # pylint: disable=no-member
    assert source_lines[0].startswith('@')
    assert source_lines[1].startswith('def ')
    assert source_lines[2][0].isspace()
    lines = str(py.code.Source('\n'.join(source_lines[2:])).deindent()).splitlines() # pylint: disable=no-member
    return lines


@contextmanager
def get_temporary_slashrc_context(new_path=None):
    prev = slash.config.root.run.user_customization_file_path
    slash.config.root.run.user_customization_file_path = new_path
    try:
        yield
    finally:
        slash.config.root.run.user_customization_file_path = prev



from types import GeneratorType
from ..exceptions import InvalidTest

from .fixtures.parameters import bound_parametrizations_context
from .fixtures.utils import nofixtures
from .requirements import get_requirements
from .runnable_test import RunnableTest
from .runnable_test_factory import RunnableTestFactory
from .tagging import get_tags


class FunctionTest(RunnableTest):

    def __init__(self, function, fixture_store, fixture_namespace, variation):
        super(FunctionTest, self).__init__(fixture_store, fixture_namespace, variation)
        self._func = function

    def get_tags(self):
        return get_tags(self._func)

    def get_address_in_factory(self):
        return ''

    def run(self):
        with bound_parametrizations_context(self._variation, self._fixture_store, self._fixture_namespace):
            result = self._fixture_store.call_with_fixtures(
                self._func, namespace=self._fixture_namespace,
                trigger_test_start=True, trigger_test_end=True,
            )
            if isinstance(result, GeneratorType):
                raise InvalidTest('{} is a generator. Running generators is not supported'.format(self._func))

    def get_test_function(self):
        return self._func

    def get_requirements(self):
        test_requirements = get_requirements(self._func)
        if nofixtures.is_marked(self._func):
            return test_requirements
        return list(set(test_requirements + self._get_fixtures_requirements()))

    def get_required_fixture_objects(self):
        return self._fixture_store.get_required_fixture_objects(self._func, namespace=self._fixture_namespace)


class FunctionTestFactory(RunnableTestFactory):

    def __init__(self, func):
        super(FunctionTestFactory, self).__init__(func)
        self.func = func

    def _generate_tests(self, fixture_store):
        namespace = fixture_store.get_current_namespace()
        for variation in fixture_store.iter_parametrization_variations(funcs=[self.func]):
            yield FunctionTest(self.func, fixture_store, namespace, variation)



import argparse
import sys

import colorama
import slash
import slash.site
from slash.utils.cli_utils import make_styler, Printer

_config_name_style = make_styler(colorama.Fore.WHITE + colorama.Style.BRIGHT) # pylint: disable=no-member
_default_style = make_styler(colorama.Fore.GREEN + colorama.Style.BRIGHT) # pylint: disable=no-member
_INDENT = ' ' * 4

def _parse_args(args):
    parser = argparse.ArgumentParser(prog='slash list-config')
    parser.add_argument('paths', nargs='*', default=[])
    parser.add_argument('--force-color', dest='force_color', action='store_true', default=False)
    parser.add_argument('--no-color', dest='enable_color', action='store_false', default=True)
    return parser.parse_args(args)


def list_config(args, report_stream=sys.stdout):

    args = _parse_args(args)
    _print = Printer(report_stream, force_color=args.force_color, enable_color=args.enable_color)

    filters = _parse_filters(args.paths)

    with slash.Session():
        slash.site.load()

        for name, value in sorted(slash.config.traverse_leaves()):
            if not _is_included(name, filters):
                continue
            _print(_config_name_style(name), '--')
            _print(_INDENT, 'default:', _default_style(value.get_value()))
            if value.metadata and 'doc' in value.metadata:
                _print(_INDENT, value.metadata['doc'])
    return 0


def _parse_filters(paths):
    returned = set()
    for p in paths:
        returned.update(_iter_subpaths(p))
    return returned

def _iter_subpaths(p):
    p = p.split('.')
    for i in range(len(p)):
        yield '.'.join(p[:i+1])


def _is_included(name, filters):
    if not filters:
        return True
    for subpath in _iter_subpaths(name):
        if subpath in filters:
            return True
    return False



# pylint: disable=redefined-outer-name,unused-variable,unused-argument
import pytest


import slash
from slash.loader import Loader

from .utils.suite_writer import Suite

_MODULE_SCOPE_ADDER = 'slash.add_cleanup({0}, scope="module")'


def test_cleanups_from_test_start(suite):

    num_tests = 5
    for _ in range(num_tests):
        suite.add_test()

    events = []
    test_ids = []

    @slash.hooks.register
    def test_start():
        test_id = slash.context.test.__slash__.id
        test_ids.append(test_id)
        events.append(('test_start', test_id))

        @slash.add_cleanup
        def cleanup():
            events.append(('test_cleanup', test_id))

    @slash.hooks.register
    def before_session_cleanup():
        events.append('before_session_cleanup')

    @slash.hooks.register
    def session_start():
        slash.add_cleanup(events.append, args=(('session_cleanup', 1),))
        slash.add_cleanup(events.append, args=(('session_cleanup', 2),))

    @slash.hooks.register
    def after_session_end():
        events.append('after_session_end')

    @slash.hooks.register
    def session_end():
        events.append('session_end')

    suite.run()

    expected_session_cleanup = [
        'before_session_cleanup',
        ('session_cleanup', 2),
        ('session_cleanup', 1),
        'session_end',
        'after_session_end',
    ]
    assert len(events) - len(expected_session_cleanup) == len(suite) * 2 == num_tests * 2

    expected = []
    for test_id in test_ids:
        expected.append(('test_start', test_id))
        expected.append(('test_cleanup', test_id))
    expected.extend(expected_session_cleanup)
    assert events == expected


def test_module_scope(scoped_suite, file1_tests, file2_tests):
    file1_test = file1_tests[0]
    file1_end = file1_tests[-1].add_deferred_event(decorator='slash.add_cleanup')
    file1_test_cleanup = file1_test.add_deferred_event(adder=_MODULE_SCOPE_ADDER)

    summary = scoped_suite.run()
    assert summary.events[file1_end].timestamp < summary.events[file1_test_cleanup].timestamp



def test_cleanups_without_session_start_never_called(checkpoint):
    assert not checkpoint.called
    with slash.Session():
        slash.add_cleanup(checkpoint)
        assert not checkpoint.called
    assert not checkpoint.called


def test_cleanups_before_session_start_get_deferred(checkpoint):
    with slash.Session() as s:
        slash.add_cleanup(checkpoint)
        with s.get_started_context():
            assert not checkpoint.called
        assert checkpoint.called_count == 1


def test_cleanups_within_cleanups_preserve_scope(checkpoint1):
    """Cleanups added from within other cleanups should happen within the scope of the parent cleanups
    """

    @slash.parametrize('x', [1, 2])
    def test_something(x):
        pass

    with slash.Session() as s:
        [fake_test1, fake_test2] = Loader().get_runnables(test_something)  # pylint: disable=unbalanced-tuple-unpacking

        s.scope_manager.begin_test(fake_test1)

        def cleanup():
            slash.add_cleanup(checkpoint1)

        slash.add_cleanup(cleanup)

        assert not checkpoint1.called
        s.scope_manager.end_test(fake_test1)

        assert checkpoint1.called


def test_errors_associated_with_correct_result(scoped_suite, file1_tests, file2_tests):
    file1_test = file1_tests[0]
    file1_test_cleanup = file1_test.add_deferred_event(adder=_MODULE_SCOPE_ADDER, extra_code=['assert 1 == 2'])
    file1_test.expect_failure()

    scoped_suite.run()


@pytest.fixture
def scoped_suite(suite, file1_tests, file2_tests):
    return suite


@pytest.fixture
def suite():
    return Suite()


@pytest.fixture
def file1_tests(suite):
    file1 = suite.add_file()
    return [file1.add_function_test() for i in range(5)]


@pytest.fixture
def file2_tests(suite):
    file2 = suite.add_file()
    return [file2.add_function_test() for i in range(3)]



import pytest
import slash
from slash._compat import izip_longest


def test_parametrization_labels(suite, suite_test):
    param = suite_test.add_parameter()
    param.add_labels()

    @suite_test.append_body
    def __code__(): # pylint: disable=unused-variable
        # pylint: disable=undefined-variablepylint: disable=no-member, protected-access, undefined-variable
        slash.context.result.data['params'] = locals().copy()
        slash.context.result.data['params'].pop('self', None)

    res = suite.run()

    for label, value, result in izip_longest(
            param.labels, param.values, res.get_all_results_for_test(suite_test)):
        assert label
        assert value
        assert result
        #assert result.test_metadata.variation.id[param.name] == label
        assert result.data['params'] == {param.name: value}
        assert label in result.test_metadata.address


def test_parametrization_labels_multiple_params(suite_builder):
    # pylint: disable=unused-variable
    @suite_builder.first_file.add_code
    def __code__():
        import slash # pylint: disable=redefined-outer-name, reimported

        @slash.parametrize(('x', 'y'),
                           [slash.param('label1', (666, 777)),
                            slash.param('label2', (888, 999)),
                           ])
        def test_1(x, y):
            slash.context.result.data['params'] = [x, y]

    suite_builder.build().run().assert_success(2).with_data([
        {'params': [666, 777]},
        {'params': [888, 999]},
    ])


def test_alternate_labeling_syntax(suite_builder):
    # pylint: disable=unused-variable
    @suite_builder.first_file.add_code
    def __code__():
        import slash # pylint: disable=reimported, redefined-outer-name

        @slash.parametrize(('x', 'y'), [
            (1, 2) // slash.param('label1'),
            (3, 4) // slash.param('label2'),
        ])
        def test_something(x, y):
            slash.context.result.data['params'] = (x, y)


    suite_builder.build().run().assert_success(2).with_data([
        {'params': (1, 2)},
        {'params': (3, 4)},
    ])


def test_must_provide_parameter_value(suite_builder):
    # pylint: disable=unused-variable,unused-argument
    @suite_builder.first_file.add_code
    def __code__():
        import slash # pylint: disable=reimported, redefined-outer-name

        @slash.parametrize('x', [
            slash.param('label1'),
        ])
        def test_something(x):
            pass

    suite_builder.build().run().assert_session_error('Parameter label1 has no value defined')


@pytest.mark.parametrize('invalid_value', [
    'x' * 31,
    '123xy',
    '123',
])
def test_invalid_param_values(invalid_value):
    with pytest.raises(RuntimeError) as caught:
        slash.param(invalid_value)
    assert 'invalid label' in str(caught.value).lower()


@pytest.mark.parametrize('valid_value', [
    'hello',
    'hello123',
    'variable_1',
    'x' * 30,
])
def test_valid_param_values(valid_value):
    slash.param(valid_value)


def test_generator_fixture_param_labels(suite_builder):
    # pylint: disable=no-member, protected-access, undefined-variable,unused-variable, reimported, redefined-outer-name
    @suite_builder.first_file.add_code
    def __code__():
        import slash

        @slash.generator_fixture
        def param():
            yield 'value1' // slash.param('first_value')
            yield 'value2' // slash.param('second_value')

        def test_1(param):
            slash.context.result.data['value'] = param

    suite_builder.build().run().assert_success(2).with_data([
        {'value': 'value1'},
        {'value': 'value2'},
    ])



# pylint: disable=unused-variable, unused-argument
import slash
import pytest

from .utils import resolve_and_run

def test_fixture_exclusion():

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.parametrize('param', [1, 2])
        @slash.fixture
        def fixture(param):
            return param * 10

        @slash.exclude('fixture.param', [2])
        def test_something(fixture):
            slash.context.result.data['value'] = fixture

        results = resolve_and_run(test_something)
    assert len(results) == 2
    assert results[0].is_success()
    assert results[0].data['value'] == 10
    assert results[1].is_skip()
    assert not results[1].is_started()


def test_fixture_exclusion_nested(suite, suite_test):

    fixture = suite.slashconf.add_fixture()
    param = fixture.add_parameter()
    suite_test.depend_on_fixture(fixture)

    param_value = param.values[0]

    suite_test.add_decorator('slash.exclude("{fixture_name}.{param_name}", {values})'.format(
        fixture_name=fixture.name,
        param_name=param.name,
        values=[param_value],
    ))
    suite_test.exclude_param_value(param.name, param_value)
    suite.run()


def test_fixture_exclusion_multiple():

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.parametrize('param', [1, 2])
        @slash.fixture
        def fixture1(param):
            return param * 10

        @s.fixture_store.add_fixture
        @slash.parametrize('param', [3, 4])
        @slash.fixture
        def fixture2(param):
            return param * 10


        @slash.exclude(('fixture1.param', 'fixture2.param'), [(2, 3)])
        def test_something(fixture1, fixture2):
            slash.context.result.data['values'] = fixture1, fixture2

        results = resolve_and_run(test_something)
    assert len(results) == 4

    value_pairs = [
        result.data.get('values')
        for result in results
    ]

    for result, value_pair in zip(results, value_pairs):
        if value_pair is None:
            assert not result.is_started()
        else:
            assert result.is_success()

    assert value_pairs == [(10, 30), (10, 40), None, (20, 40)]


@pytest.mark.parametrize('names,values', [
    (['x'], 1),
    (['x'], [1]),
])
def test_invalid_exclude_patterns(names, values):
    with pytest.raises(RuntimeError) as caught:
        slash.exclude(names, values)
    assert 'Invalid exclude values' in str(caught.value)

@pytest.mark.parametrize('names,values', [
    ('x', [1]),
    (['x'], [(1,)]),
])
def test_valid_exclude_patterns(names, values):
    slash.exclude(names, values)



# pylint: disable=redefined-outer-name
from uuid import uuid4

import gossip

import pytest


def test_facts(suite, suite_test, fact_name, fact_value):

    facts = []

    @gossip.register('slash.fact_set')
    def fact_set(name, value): # pylint: disable=unused-variable
        facts.append((name, value))

    suite_test.append_line('slash.context.result.facts.set({!r}, {!r})'.format(fact_name, fact_value))
    res = suite.run()[suite_test]
    assert res.facts.all() == {fact_name : fact_value}
    assert facts == [(fact_name, fact_value)]


@pytest.fixture
def fact_name():
    return 'some_fact'


@pytest.fixture
def fact_value():
    return str(uuid4())



from ..interface import PluginInterface
from ... import hooks
from ...conf import config
from ...ctx import session
from ...exception_handling import handling_exceptions
from ...utils.conf_utils import Cmdline
from slash.plugins import parallel_mode
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from vintage import warn_deprecation
import datetime
from pkg_resources import resource_string
import requests
import slash
import smtplib
import jinja2


_SLASH_ICON = "http://slash.readthedocs.org/en/latest/_static/slash-logo.png"

def _post_request(url, **kwargs):
    response = requests.post(url, **kwargs)
    response.raise_for_status()
    return response


def _send_email(smtp_server, subject, body, from_email, to_list, cc_list):
    """Send an email.

    :param str smtp_server: The smtp_server uri
    :param str subject: The email subject.
    :param str body: The email body.
    :param str from_email: The from address.
    :param list to_list: A list of email addresses to send to.
    :param list cc_list: A list of email addresses to send to as cc.
    """
    msg = MIMEMultipart('alternative')
    msg.attach(MIMEText(body, 'html'))

    if not to_list:
        to_list = []

    if not cc_list:
        cc_list = []

    msg['Subject'] = subject
    msg['From'] = from_email
    msg['To'] = ', '.join(to_list)
    msg['Cc'] = ', '.join(cc_list)
    smtp = smtplib.SMTP(smtp_server, timeout=30)
    try:
        smtp.sendmail(from_email, to_list + cc_list, msg.as_string())
    finally:
        smtp.quit()


class Message(object):
    def __init__(self, title, body_template, kwargs, is_pdb):
        self.title = title
        self.body = body_template
        self.kwargs = kwargs
        self.details_dict = {}
        self.is_pdb = is_pdb

    def get_title(self):
        return self.title.format(**self.kwargs)

    def get_short_message(self):
        return self.body.format(**self.kwargs)

    _email_template = None

    def _get_html_template(self):
        returned = self._email_template
        if returned is None:
            source = resource_string('slash.plugins.builtin', 'email_template.j2').decode('utf-8')
            returned = type(self)._email_template = jinja2.Template(source)
        return returned

    def get_html_message(self):
        return self._get_html_template().render(message=self, **self.kwargs)


@parallel_mode('parent-only')
class Plugin(PluginInterface):
    """Enables notifications for successful and failed test runs through various providers (NMA, Prowl, Pushover etc.)
    For more information see https://slash.readthedocs.org/en/master/builtin_plugins.html#notifications
    """

    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self._notifiers = {}
        self._basic_config = {
            "prowl_api_key" : None,
            "nma_api_key" : None,
            "pushbullet_api_key": None,
            "notification_threshold": 5,
            "notify_only_on_failures": False // Cmdline(on='--notify-only-on-failures'),
            "notify_on_pdb": True,
        }
        self._add_notifier(self._prowl_notifier, 'prowl', {'api_key': None, 'enabled': True})
        self._add_notifier(self._nma_notifier, 'nma', {'api_key': None, 'enabled': True})
        self._add_notifier(self._pushbullet_notifier, 'pushbullet', {'api_key': None, 'enabled': True})
        self._add_notifier(self._email_notifier, 'email', {
            'from_email': 'Slash <noreply@getslash.github.io>',
            'smtp_server': None,
            'to_list': [] // Cmdline(append='--email-to', metavar='ADDRESS'),
            'cc_list': []
        })
        self._add_notifier(self._slack_notifier, 'slack', {'url': None, 'channel': None, 'from_user': 'slash-bot'})

    def get_name(self):
        return 'notifications'

    def get_default_config(self):
        return self._basic_config

    def _add_notifier(self, func, name, conf_dict=None):
        self._notifiers[name] = func
        if not conf_dict:
            conf_dict = {}
        assert isinstance(conf_dict, dict)
        conf_dict.setdefault('enabled', False)
        conf_dict['enabled'] //= Cmdline(on='--notify-{}'.format(name))
        self._basic_config[name] = conf_dict

    def _get_from_config_with_legacy(self, notifier_name, legacy_name, new_name):
        this_config = config.get_path('plugin_config.notifications')
        value = this_config[legacy_name]
        if value:
            warn_deprecation('{} is depreacted. use {}.{} instead'.format(legacy_name, notifier_name, new_name))
        else:
            value = this_config[notifier_name][new_name]
        return value

    @staticmethod
    def _os_post_notification(url, api_key, message):
        if api_key:
            data = {
                "apikey": api_key,
                "application": "slash",
                "event": message.get_title(),
                "description": message.get_short_message(),
            }
            _post_request(url, data=data)

    def _prowl_notifier(self, message):
        api_key = self._get_from_config_with_legacy('prowl', 'prowl_api_key', 'api_key')
        self._os_post_notification("https://prowl.weks.net/publicapi/add", api_key, message)

    def _nma_notifier(self, message):
        api_key = self._get_from_config_with_legacy('nma', 'nma_api_key', 'api_key')
        self._os_post_notification("https://www.notifymyandroid.com/publicapi/notify", api_key, message)

    def _pushbullet_notifier(self, message):
        api_key = self._get_from_config_with_legacy('pushbullet', 'pushbullet_api_key', 'api_key')
        if api_key:
            data = {"type": "note", "title": message.get_title(), "body": message.get_short_message()}
            _post_request("https://api.pushbullet.com/api/pushes", data=data, auth=(api_key, ""))

    def _email_notifier(self, message):
        email_config = config.root.plugin_config.notifications.email
        email_kwargs = {
            'from_email': email_config.from_email,
            'subject': message.get_title(),
            'body': message.get_html_message(),
            'smtp_server': email_config.smtp_server,
            'to_list': email_config.to_list or None,
            'cc_list': email_config.cc_list,
        }
        if all(value is not None for value in email_kwargs.values()):
            _send_email(**email_kwargs)

    def _slack_notifier(self, message):
        slack_config = config.root.plugin_config.notifications.slack
        if (slack_config.url is None) or (slack_config.channel is None):
            return

        color = '#439FE0' if message.is_pdb else ('good' if self._finished_successfully() else 'danger')
        kwargs = {
            'attachments': [{
                'title': message.get_title(),
                'fallback': 'Session {session_id} {result}'.format(**message.kwargs),
                'text': message.get_short_message(),
                'color': color,
                }],
            'channel': slack_config.channel,
            'username': slack_config.from_user,
            'icon_url': _SLASH_ICON,
            }
        _post_request(slack_config.url, json=kwargs)

    def _finished_successfully(self):
        return session.results.is_success(allow_skips=True)

    def _get_message(self, short_message, is_pdb):
        result_str = 'entered PDB' if is_pdb else ("Succeeded" if self._finished_successfully() else "Failed")
        kwargs = {
            'session_id': session.id,
            'host_name': session.host_name,
            'full_name': 'N/A',
            'duration': str(datetime.timedelta(seconds=session.duration)).partition('.')[0],
            'result': result_str,
            'success': self._finished_successfully(),
            'results_summary': repr(session.results).replace('<', '').replace('>', ''),
            'total_num_tests': session.results.get_num_results(),
            'non_successful_tests': session.results.get_num_errors() + session.results.get_num_failures(),
        }
        backslash_plugin = slash.plugins.manager.get_active_plugins().get('backslash')
        if backslash_plugin:
            config.root.plugin_config.notifications.email.to_list.append(backslash_plugin.session.user_email)
            url = backslash_plugin.webapp_url + 'sessions/{}'.format(session.id)
            kwargs['backslash_link'] = url
            kwargs['full_name'] = backslash_plugin.session.user_display_name
            session_info = 'Backslash: {backslash_link}'
        else:
            session_info = 'Session ID: {session_id}'
        short_message += "\n\n" + session_info
        title = "Slash Session on {host_name} has {result}".format(**kwargs)
        kwargs['title'] = title
        return Message(title, short_message, kwargs, is_pdb)

    def entering_debugger(self, exc_info):
        if not self.current_config.notify_on_pdb:
            return
        self._notify_all(short_message=repr(exc_info[1]), is_pdb=True)

    def session_end(self):
        if session.duration < self.current_config.notification_threshold:
            return
        if self._finished_successfully() and self.current_config.notify_only_on_failures:
            return
        self._notify_all(short_message='{results_summary}', is_pdb=False)


    def _notify_all(self, short_message, is_pdb):
        message = self._get_message(short_message, is_pdb)
        hooks.prepare_notification(message=message) # pylint: disable=no-member
        this_config = config.get_path('plugin_config.notifications')

        for notifier_name, notifier_func in self._notifiers.items():
            if not this_config[notifier_name]['enabled']:
                continue
            with handling_exceptions(swallow=True):
                notifier_func(message)



# pylint: disable=redefined-outer-name
import pytest
import slash
from slash import ctx, context, Session
from slash.ctx import Context, ContextAttributeProxy
from slash.reporting.null_reporter import NullReporter


@pytest.mark.parametrize('on_stack', [True, False])
def test_context_dir_method_loaded_context(loaded_context, on_stack):
    output = dir(context if on_stack else loaded_context)
    assert 'test_id' in output


def test_context_dir_no_context_loaded():
    assert 'test_id' in dir(context)


def test_null_context_cant_setattr():
    with pytest.raises(AttributeError):
        context.x = 2
    assert not hasattr(context, 'x')


def test_dir_object(loaded_context, queried_context):
    class Object(object):
        x = 2
        y = 3
    loaded_context.test_id = obj = Object()
    assert dir(queried_context.test_id) == dir(obj)


def test_call_object(loaded_context, queried_context):
    value = 'some value'

    def func():
        return value
    loaded_context.test_id = func

    assert queried_context.test_id() == value


def test_null_context():
    assert context.test_id is None
    assert context.result is None


def test_no_session_session_id(loaded_context):
    assert loaded_context.session_id is None


def test_no_session_reporter(loaded_context):
    assert isinstance(loaded_context.reporter, NullReporter)


def test_no_test_test_filename(loaded_context):
    assert loaded_context.test_filename is None


def test_session_context_result():
    with Session() as s:
        assert context.session is s
        assert context.result is s.results.global_result


def test_context_test_filename(suite, suite_test):
    suite_test.append_line('assert slash.context.test_filename == slash.context.test.__slash__.file_path')
    suite.run()


def test_test_context_result(suite):
    for test in suite:
        test.append_line('assert slash.context.result is slash.session.results[slash.context.test]')

    @slash.hooks.result_summary.register  # pylint: disable=no-member
    def assert_result_back_to_normal():  # pylint: disable=unused-variable
        assert context.result is context.session.results.global_result

    assert suite.run().ok()


def test_cannot_pop_bottom():
    assert len(context._stack) == 1  # pylint: disable=protected-access
    with pytest.raises(RuntimeError):
        context.pop()


def test_push_context(loaded_context):
    test_id = loaded_context.test_id = "some_test_id"
    assert context.test_id == test_id


def test_object_proxy_getattr(contextobj, realobj):
    assert contextobj.attr is realobj.attr
    assert contextobj.__attr__ is realobj.__attr__
    assert not hasattr(contextobj, "nonexisting")
    assert not hasattr(contextobj, "__nonexisting__")


def test_object_proxy_eq(contextobj, realobj):
    assert contextobj == realobj


def test_object_proxy_ne(contextobj, realobj):
    assert not (contextobj != realobj)  # pylint: disable=superfluous-parens


def test_object_proxy_str_repr(contextobj, realobj):
    assert str(contextobj) == str(realobj)
    assert repr(contextobj) == repr(realobj)


@pytest.fixture
def contextobj(loaded_context, realobj):
    setattr(loaded_context, "obj", realobj)
    return ContextAttributeProxy("obj")


@pytest.fixture
def realobj():
    class Object(object):

        __attr__ = object()
        attr = object()

    return Object()


@pytest.fixture
def loaded_context():
    returned = Context()
    context.push(returned)
    return returned


@pytest.fixture(autouse=True, scope="function")
def pop_all(request):
    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        while len(context._stack) > 1:  # pylint: disable=protected-access
            context.pop()

@pytest.fixture(params=['global', 'proxy'])
def queried_context(request):
    if request.param == 'global':
        return context
    elif request.param == 'proxy':
        return ctx
    else:
        raise NotImplementedError() # pragma: no cover



# pylint: disable=pointless-statement,redefined-outer-name
import pytest
import slash


def test_frame_order(error, use_exception):
    code_line = error.traceback.frames[-1].code_line.strip()
    if use_exception:
        assert code_line == '1 / 0'
    else:
        assert code_line == 'get_error_adder()("error message")'


def test_self_variables(error):
    frame = error.traceback.frames[-3]
    assert frame.func_name == 'method1'
    assert 'ExampleObject' in frame.locals['self']['value']
    assert frame.locals['self.a']['value'] == '1'
    assert frame.locals['self.b']['value'] == '2'
    for var_name in frame.locals:
        assert not var_name.startswith('self.__')


@pytest.fixture
def error(get_error_adder, use_exception):

    def f1():
        obj = ExampleObject()
        try:
            obj.method1()
        except ZeroDivisionError:
            assert use_exception
            get_error_adder()()



    class ExampleObject(object):

        def __init__(self):
            self.a = 1
            self.b = 2

        def method1(self):
            g1()


    def g1():
        h1()

    def h1():
        if use_exception:
            1 / 0
        else:
            get_error_adder()("error message")

    with slash.Session() as s:
        f1()
    [err] = s.results.global_result.get_errors()
    return err


@pytest.fixture(params=[lambda: slash.add_error, lambda: slash.context.result.add_error])
def get_error_adder(request):
    return request.param


@pytest.fixture(params=[True, False])
def use_exception(request):
    return request.param



# pylint: disable=redefined-outer-name
import pytest
from slash._compat import StringIO
from slash.frontend.slash_list_plugins import slash_list_plugins
from slash.plugins import manager, PluginInterface


def test_slash_list_plugins(report_stream):
    slash_list_plugins([], report_stream=report_stream)
    output = report_stream.getvalue()
    assert output
    installed = manager.get_installed_plugins()
    for plugin_name in installed:
        assert plugin_name in output


def test_slash_list_plugins_for_internal_plugins(report_stream):
    internal_plugin = InternalPlugin()
    manager.install(internal_plugin, is_internal=True)
    slash_list_plugins([], report_stream=report_stream)
    output = report_stream.getvalue()
    assert output

    assert internal_plugin.get_name() in manager.get_installed_plugins()
    assert internal_plugin.get_name() not in output
    assert '--internal-plugin-option' not in output


@pytest.fixture
def report_stream():
    return StringIO()


class InternalPlugin(PluginInterface):

    def get_name(self):
        return "internal-plugin"

    def configure_argument_parser(self, parser):
        parser.add_argument("--internal-plugin-option")

    def configure_from_parsed_args(self, args):
        self.cmdline_param = args.plugin_option



from ..ctx import context
from ..exceptions import UnknownFixtures
from .fixtures.parameters import Parametrization
from . import markers


def exclude(names, values):
    """
    Excludes a specific parametrization of a test from running

    :param name: can receive either a name of a parameter for this test, or a name of a fixture parameter
    :param values: must be a list of values to exclude for the given parameter
    """
    if not isinstance(values, (tuple, list)):
        raise RuntimeError('Invalid exclude values specified: must be a sequence, got {!r}'.format(values))
    if not isinstance(names, (tuple, list)):
        names = (names,)
        values = [(value,) for value in values]
    elif not isinstance(values, (tuple, list)) or any(not isinstance(item, tuple) for item in values):
        raise RuntimeError('Invalid exclude values specified for {}: {!r}'.format(', '.join(names), values))

    values = [tuple(value_set) for value_set in values]
    return markers.exclude_marker((names, values))

def is_excluded(test):
    test_func = test.get_test_function()
    exclusions = markers.exclude_marker.get_value(test_func, default=None)
    if not exclusions:
        return False
    exclusions = dict(exclusions)
    for parameter_names, value_sets in exclusions.items():
        params = []
        values = []
        for parameter_name in parameter_names:
            param = context.session.fixture_store.resolve_name(parameter_name, start_point=test_func, namespace=test.get_fixture_namespace())
            if not isinstance(param, Parametrization):
                raise UnknownFixtures('{!r} is not a parameter, and therefore cannot be the base for value exclusions'.format(parameter_name))
            params.append(param)


            try:
                param_index = test.__slash__.variation.param_value_indices[param.info.id] #pylint: disable=no-member
            except LookupError:
                raise UnknownFixtures('{!r} cannot be excluded for {!r}'.format(parameter_name, test))
            value = param.get_value_by_index(param_index)
            values.append(value)

        if tuple(values) in value_sets:
            return True
    return False



# pylint: disable=unused-variable, unused-argument, redefined-outer-name
import collections
import copy
from uuid import uuid4

import gossip
import pytest
import slash
from munch import Munch

from .utils import make_runnable_tests


def test_variation_info_without_params_is_none(results):
    assert len(results.test_no_params) == 1
    for res in results.test_no_params:
        assert res.test_metadata.variation.values == {}


def test_variation_info_single_value_id_none(results):
    assert len(results.test_single_param_fixture) == 1
    for res in results.test_single_param_fixture:
        assert res.test_metadata.variation.id is not None
        assert res.test_metadata.variation.values == {}
        assert 'fixture' not in res.data['captured_values']

def test_unique_variation_ids(results):
    all_results = [res for result_set in results.values() for res in result_set]
    ids = {_freeze(res.test_metadata.variation.id) for res in all_results}
    assert len(ids) == len(all_results) - 1 # we subtract one because no params and a single fixture have the same id
    assert None not in ids


def test_parametrization_info_availabe_on_test_start(checkpoint):
    param_value = str(uuid4())

    @gossip.register('slash.test_start')
    def test_start_hook():
        assert slash.context.test.__slash__.variation.values['param'] == param_value
        assert 'fixture' not in slash.context.test.__slash__.variation.values
        checkpoint()

    @slash.parametrize('param', [param_value])
    def test_something(param, fixture):
        pass

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        def fixture():
            pass

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))

    assert s.results.is_success(allow_skips=False)
    assert checkpoint.called



def test_parametrization_info_values_include_nested_fixture_values():

    value1 = str(uuid4())
    value2 = str(uuid4())

    @gossip.register('slash.test_start')
    def test_start_hook():
        slash.context.result.data['variation_values'] = slash.context.test.__slash__.variation.values.copy()

    @slash.parametrize('param', ['some_value'])
    def test_something(param, some_fixture):
        pass

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        @slash.parametrize('value', [value1, value2])
        def some_fixture(value):
            pass

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))

    assert s.results.is_success(allow_skips=False)
    all_values = []
    for result in s.results.iter_test_results():
        values = result.data['variation_values']
        all_values.append(values['some_fixture.value'])

    assert len(all_values) == 2
    assert set(all_values) == {value1, value2}


def test_variation_identification():
    value1 = str(uuid4())
    value2 = str(uuid4())

    @gossip.register('slash.test_start')
    def test_start_hook():
        variation = slash.context.test.__slash__.variation
        slash.context.result.data['variation_info'] = {
            'id': variation.id.copy(),
            'values': variation.values.copy(),
        }


    @slash.parametrize('param', ['some_value'])
    def test_something(param, some_fixture):
        pass

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        @slash.parametrize('value', [value1])
        def some_fixture(value):
            return value2

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))

    assert s.results.is_success(allow_skips=False)
    [info] = [result.data['variation_info'] for result in s.results.iter_test_results()]
    assert info['id']['param'] == 0
    assert info['values']['param'] == 'some_value'
    assert info['id']['some_fixture.value'] == 0
    assert 'some_fixture' not in info['values']
    assert info['values']['some_fixture.value'] == value1





def _freeze(dictionary):
    return frozenset(dictionary.items())

def test_variation_tuples(results):
    [res] = results.test_parametrization_tuple
    values = res.data['captured_values']
    assert values['x'] == 1
    assert values['y'] == 2

def test_nested_fixture_ids(results):
    ids = {res.data['captured_values']['outer_fixture.outer_param'] for res in results.test_nested_fixture}
    assert ids == {666}
    for res in results.test_nested_fixture:
        assert 'outer_fixture' not in res.data['captured_values']

def test_fixture_and_toggle(results):
    assert len(results.test_fixture_and_toggle) == 2


@pytest.fixture
def results():

    tests = []
    def include(f):
        tests.append(f)
        return f

    @include
    def test_no_params():
        pass

    @include
    def test_single_param_fixture(fixture):
        _capture_arguments()

    @include
    def test_nested_fixture(outer_fixture):
        _capture_arguments()

    @include
    @slash.parametrize(('x', 'y'), [(1, 2)])
    def test_parametrization_tuple(x, y):
        _capture_arguments()

    @include
    @slash.parameters.toggle('toggle')
    def test_fixture_and_toggle(fixture, toggle):
        _capture_arguments()

    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.fixture
        def fixture():
            return _object1

        @s.fixture_store.add_fixture
        @slash.fixture
        @slash.parametrize('x', [1, 2, 3])
        def inner_fixture(x):
            return 'inner{}'.format(x)

        @s.fixture_store.add_fixture
        @slash.fixture
        @slash.parametrize('outer_param', [666])
        def outer_fixture(inner_fixture, outer_param):
            return 'outer_{}'.format(inner_fixture)

        s.fixture_store.resolve()

        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(tests))
    assert s.results.is_success(allow_skips=False)

    returned = collections.defaultdict(list)
    for res in s.results.iter_test_results():
        returned[res.test_metadata.function_name].append(res)
    return Munch(returned)

# helpers ################################################################################

_object1 = object()


def _capture_arguments():
    values = copy.copy(slash.context.result.test_metadata.variation.values)
    slash.context.result.data['captured_values'] = values



import operator
import re
from contextlib import contextmanager
from sentinels import NOTHING

from ..._compat import iteritems
from ...exception_handling import mark_exception_frame_correction
from ...utils.python import wraps, get_argument_names
from .fixture_base import FixtureBase
from .utils import FixtureInfo, get_scope_by_name

_PARAM_INFO_ATTR_NAME = '__slash_parametrize__'


def parametrize(parameter_name, values):
    """Decorator to create multiple test cases out of a single function or module, where the cases vary by the value of ``parameter_name``,
    as iterated through ``values``.
    """

    def decorator(func):

        params = getattr(func, _PARAM_INFO_ATTR_NAME, None)
        if params is None:
            params = ParameterizationInfo(func)

            @wraps(func, preserve=['__slash_fixture__'])
            def new_func(*args, **kwargs):
                # for better debugging. _current_variation gets set to None on context exit
                variation = _current_variation
                for name, param in params.iter_parametrization_fixtures():
                    value = variation.get_param_value(param)
                    if name not in kwargs:
                        kwargs[name] = value
                return func(*args, **kwargs)
            setattr(new_func, _PARAM_INFO_ATTR_NAME, params)
            returned = new_func
        else:
            returned = func

        params.add_options(parameter_name, values)
        return returned

    return decorator


def iterate(**kwargs):

    def decorator(func):
        for name, options in iteritems(kwargs):
            func = parametrize(name, options)(func)
        return func
    return decorator


def toggle(param_name):
    """A shortcut for :func:`slash.parametrize(param_name, [True, False]) <slash.parametrize>`

    .. note:: Also available for import as slash.parameters.toggle
    """
    return parametrize(param_name, (True, False))


_current_variation = None


@contextmanager
def bound_parametrizations_context(variation, fixture_store, fixture_namespace):
    global _current_variation  # pylint: disable=global-statement
    assert _current_variation is None
    _current_variation = variation
    try:
        fixture_store.activate_autouse_fixtures_in_namespace(fixture_namespace)
        yield
    finally:
        _current_variation = None


def iter_parametrization_fixtures(func):
    if isinstance(func, FixtureBase):
        func = func.fixture_func
    param_info = getattr(func, _PARAM_INFO_ATTR_NAME, None)
    if param_info is None:
        return []
    return param_info.iter_parametrization_fixtures()


class ParameterizationInfo(object):

    def __init__(self, func):
        super(ParameterizationInfo, self).__init__()
        self._argument_names = get_argument_names(func)
        self._argument_name_set = frozenset(self._argument_names)
        self._params = {}
        self._extra_params = {}
        self.path = '{}:{}'.format(func.__module__, func.__name__)

    def add_options(self, param_name, values):
        assert param_name not in self._params
        values = list(values)

        if not isinstance(param_name, (list, tuple)):
            names = (param_name,)
        else:
            names = param_name

        values = _normalize_values(values, num_params=len(names))

        p = Parametrization(values=values, path='{}.{}'.format(self.path, param_name))
        for index, name in enumerate(names):
            if name in self._argument_name_set:
                params_dict = self._params
            else:
                params_dict = self._extra_params
            if len(names) > 1:
                params_dict[name] = p.as_transform(operator.itemgetter(index))
            else:
                params_dict[name] = p

    def iter_parametrization_fixtures(self):
        for name in self._argument_names:
            values = self._params.get(name)
            if values is not None:
                yield name, values
        for name, values in self._extra_params.items():
            yield name, values


def _id(obj):
    return obj


class Parametrization(FixtureBase):

    def __init__(self, path, values, info=None, transform=_id):
        super(Parametrization, self).__init__()
        self.path = path
        self.values = values
        if info is None:
            info = FixtureInfo(path=path)
        self.info = info
        self.scope = get_scope_by_name('test')
        self.transform = transform

    def get_value_by_index(self, index):
        return self.transform(self._compute_value(self.values[index]))

    def _compute_value(self, param):
        if isinstance(param, list):
            return [p.value for p in param]
        return param.value

    def is_parameter(self):
        return True

    def is_fixture(self):
        return False

    def get_value(self, kwargs, active_fixture):
        raise NotImplementedError()  # pragma: no cover

    def get_variations(self):
        raise NotImplementedError()  # pragma: no cover

    def _resolve(self, store):
        return {}

    def as_transform(self, transform):
        returned = Parametrization(values=self.values, info=self.info, transform=transform, path=self.path)
        return returned


class ParametrizationValue(object):

    def __init__(self, label, value=NOTHING):
        super(ParametrizationValue, self).__init__()
        self._validate_label(label)
        self.label = label
        self.value = value

    def _validate_label(self, label):
        if isinstance(label, str) and not re.match(r'^[a-zA-Z_][0-9a-zA-Z_]{0,29}$', label):
            raise RuntimeError('Invalid label: {!r}'.format(label))

    def __rfloordiv__(self, other):
        assert self.value is NOTHING, 'Parameter already has a value'
        self.value = other
        return self


def _normalize_values(values, num_params=1):
    returned = []
    for index, value in enumerate(values):

        value = _normalize_single_value(value, default_label=index)
        if num_params > 1:
            if not isinstance(value.value, (tuple, list)):
                raise RuntimeError('Invalid parametrization value (expected sequence): {!r}'.format(value.value))
            if len(value.value) != num_params:
                raise RuntimeError('Invalid parametrization value (invalid length, expected {}): {!r}'.format(num_params, value.value))

        returned.append(value)
    return returned


def _normalize_single_value(value, default_label):
    if not isinstance(value, ParametrizationValue):
        value = ParametrizationValue(label=default_label, value=value)
    if value.value is NOTHING:
        raise mark_exception_frame_correction(
            RuntimeError('Parameter {} has no value defined!'.format(value.label)),
            +4)
    return value



# pylint: disable=line-too-long
import logbook
from confetti import Config
from .utils.conf_utils import Doc, Cmdline

__all__ = ["config"]

config = Config({
    "debug": {
        "debug_skips": False // Doc("Enter pdb also for SkipTest exceptions"),
        "debug_hook_handlers": False // Doc("Enter pdb also for every exception encountered in a hook/callback. Only relevant when debugging is enabled"),
        "enabled": False // Doc("Enter pdb on failures and errors") // Cmdline(on="--pdb"),
        "filter_strings": [] // Doc("A string filter, selecting if to enter pdb") // Cmdline(append='--pdb-filter',
                                                                                             metavar='FILTER'),
        "debugger": None,
    },

    "log": {
        "colorize": False // Doc("Emit log colors to files"),
        "console_theme": {
            'inline-file-end-fail': 'red',
            'inline-file-end-skip': 'yellow',
            'inline-file-end-success': 'green',
            'inline-error': 'red',
            'inline-test-interrupted': 'yellow',
            'error-cause-marker': 'white/bold',
            'fancy-message': 'yellow/bold',
            'frame-local-varname': 'yellow/bold',
            'session-summary-success': 'green/bold',
            'session-summary-failure': 'red/bold',
            'error-separator-dash': 'red',
            'tb-error-message': 'red/bold',
            'tb-error': 'red/bold',
            'tb-frame-location': 'white/bold',
            'test-additional-details-header': 'black/bold',
            'test-additional-details': 'black/bold',
            'test-error-header': 'white',
            'test-skip-message': 'yellow',
            'tb-line-cause': 'white',
            'tb-test-line': 'red/bold',
            'tb-line': 'black/bold',
        },

        "console_level": logbook.WARNING // Cmdline(decrease='-v',
                                                    decrease_doc='Make console more verbose (can be specified multiple times)',
                                                    increase='-q',
                                                    increase_doc='Make console less verbose (can be specified multiple times)'),
        "core_log_level": logbook.WARNING // Doc("Minimal level of slash log messages to show"),
        "color_console": None // Cmdline(on='--force-color', off='--no-color'),
        "traceback_variables": False // Doc("Logs values of variables in traceback frames for added errors"),
        "console_traceback_level": 2 // Doc("Detail level of tracebacks") // Cmdline(arg="--tb"),
        "truncate_console_lines": True // Doc("truncate long log lines on the console") // Cmdline(arg='--truncate-console-lines', metavar='yes/no'),
        "truncate_console_errors": False // Doc("If truncate_console_lines is set, also truncate long log lines, including and above the \"error\" level, on the console"),
        "root": None // Doc("Root directory for logs") // Cmdline(arg="-l", metavar="DIR"),
        "subpath": "{context.session.id}/{context.test_id}/debug.log" // Doc("Path to write logs to under the root"),
        "session_subpath": "{context.session.id}/session.log",
        "errors_subpath": None // Doc("Depreacted - Use 'highlights_subpath' config instead"),
        "highlights_subpath": None // Doc("If set, this path will be used to record highlights (eg. errors added) in the session and/or tests"),
        "last_session_symlink": None // Doc("If set, specifies a symlink path to the last session log file in each run"),
        "last_session_dir_symlink": None // Doc("If set, specifies a symlink path to the last session log directory"),
        "last_test_symlink": None // Doc("If set, specifies a symlink path to the last test log file in each run"),
        "last_failed_symlink": None // Doc("If set, specifies a symlink path to the last failed test log file"),
        "show_manual_errors_tb": True // Doc("Show tracebacks for errors added via slash.add_error"),
        "show_raw_param_values": False // Doc("Makes test start logs contain the raw values of test parameters"),

        "silence_loggers": [] // Doc("Logger names to silence"),
        "format": None // Doc("Format of the log line, as passed on to logbook. None will use the default format"),
        "console_format": None // Doc("Optional format to be used for console output. Defaults to the regular format"),
        "localtime": False // Doc("Use local time for logging. If False, will use UTC"),
        "unittest_mode": False // Doc("Used during unit testing. Emit all logs to stderr as well as the log files"),
        "unified_session_log": False // Doc("Make the session log file contain all logs, including from tests"),
        "compression": {
            "enabled": False // Doc("Compress log files"),
            "algorithm": "brotli" // Doc("Compression algorithm to use, either gzip or brotli"),
            "use_rotating_raw_file": False // Doc("When compression is enabled, write also to uncompressed rotating log file"),
        },
        "cleanup": {
            "enabled": False,
            "keep_failed": True,
        }
    },
    "run": {
        "dump_variation": False // Doc("Output the full variation structure before each test is run (mainly used for internal debugging)"),
        "default_sources": [] // Doc("Default tests to run assuming no other sources are given to the runner"),
        "suite_files": [] // Doc("File(s) to be read for lists of tests to be run") // Cmdline(append="-f", metavar="FILENAME"),
        "stop_on_error": False // Doc("Stop execution when a test doesn't succeed") // Cmdline(on="-x", off="-X"),
        "filter_strings": [] // Doc("A string filter, selecting specific tests by string matching against their name") // Cmdline(append='-k', metavar='FILTER'),
        "repeat_each": 1 // Doc("Repeat each test a specified amount of times") // Cmdline(arg='--repeat-each', metavar="NUM_TIMES"),
        "repeat_all": 1 // Doc("Repeat all suite a specified amount of times") // Cmdline(arg='--repeat-all', metavar="NUM_TIMES"),
        "session_state_path": "~/.slash/last_session" // Doc("Where to keep last session serialized data"),
        "project_name": None,
        "project_customization_file_path": "./.slashrc",
        "user_customization_file_path": "~/.slash/slashrc",
        "resume_state_path": "~/.slash/session_states" // Doc("Path to store or load session's resume data"),
        "message_assertion_introspection": True // Doc("When False, failing assertions which have messages attached will not emit introspection info"),
    },
    "interactive": {
        "expose_g_globals": True // Doc("When False, slash.g won't be added to interactive test namespaces"),
    },

    "parallel": {
        "num_workers": 0 // Doc("Parallel execution") // Cmdline(arg='--parallel', metavar="NUM_WORKERS"),
        "worker_id": None // Doc("Worker_id") // Cmdline(arg='--parallel-worker-id', metavar="WORKER_ID"),
        "server_addr": "localhost" // Doc("Server address") // Cmdline(arg='--parallel-addr', metavar="PARALLEL_SERVER_ADDRESS"),
        "server_port": 0 // Doc("Server port") // Cmdline(arg='--parallel-port', metavar="PARALLEL_SERVER_PORT"),
        "parent_session_id": None // Doc("parent session id") // Cmdline(arg='--parallel-parent-session-id', metavar="MASTER_SESSION_ID"),
        "communication_timeout_secs": 60 // Doc("timeout of worker in seconds"),
        "worker_connect_timeout": 10 // Doc("timeout for each worker to connect"),
        "no_request_timeout": 20 // Doc("timeout for server not getting requests"),
        "worker_error_file": "errors-worker" // Doc("worker error filename template"),
        "workers_error_dir": None // Doc("workers error directory") // Cmdline(arg='--workers-error-dir', metavar="WORKERS_ERROR_DIR"),
    },
    "tmux": {
        "enabled": False // Doc("Run inside tmux") // Cmdline(on="--tmux"),
        "use_panes": False // Doc("In parallel mode, run children inside panes and not windows") // Cmdline(on="--tmux-panes"),
    },
    "sentry": {
        "dsn": None // Doc("Possible DSN for a sentry service to log swallowed exceptions. "
                           "See http://getsentry.com for details"),
    },
    "plugins": {
        "search_paths": [] // Doc("List of paths in which to search for plugin modules"),
    },

    "plugin_config": {
        # DO NOT store configuration here. It is intended for dynamically loaded plugins
    },
})



class RegistrationInfo(object):

    def __init__(self, hook_name, expect_exists, register_kwargs=None):
        assert isinstance(expect_exists, bool)
        self.hook_name = hook_name
        self.expect_exists = expect_exists
        self.register_kwargs = register_kwargs or {}



# pylint: disable=superfluous-parens,protected-access
from slash.utils.marks import mark, try_get_mark
from slash import Session
from slash.utils import parallel_utils
import pytest
import slash.plugins


@pytest.mark.parametrize('session_state', ['not_parallel', 'parent', 'child'])
@pytest.mark.parametrize('plugin_parallel_mode', parallel_utils.ParallelPluginModes.MODES)
def test_parallel_mode(plugin, config_override, session_state, plugin_parallel_mode):
    slash.plugins.manager._pending_deactivation = set()

    if not session_state == 'not_parallel':
        config_override("parallel.num_workers", 1)
    if session_state == 'child':
        config_override("parallel.worker_id", 1)

    if session_state == 'not_parallel':
        assert not parallel_utils.is_parallel_session()
        assert not parallel_utils.is_parent_session()
        assert not parallel_utils.is_child_session()
    if session_state == 'parent':
        assert parallel_utils.is_parallel_session()
        assert parallel_utils.is_parent_session()
        assert not parallel_utils.is_child_session()
    if session_state == 'child':
        assert parallel_utils.is_parallel_session()
        assert parallel_utils.is_child_session()
        assert not parallel_utils.is_parent_session()

    plugin = mark('parallel_mode', plugin_parallel_mode)(plugin)
    assert try_get_mark(plugin, 'parallel_mode') == plugin_parallel_mode

    slash.plugins.manager.install(plugin)
    slash.plugins.manager.configure_for_parallel_mode()
    is_plugin_in_deactivation_list = plugin.get_name() in slash.plugins.manager._pending_deactivation
    if plugin_parallel_mode == parallel_utils.ParallelPluginModes.ENABLED:
        assert not is_plugin_in_deactivation_list
    elif plugin_parallel_mode == parallel_utils.ParallelPluginModes.DISABLED:
        assert is_plugin_in_deactivation_list if session_state != "not_parallel" else not is_plugin_in_deactivation_list
    elif plugin_parallel_mode == parallel_utils.ParallelPluginModes.PARENT_ONLY:
        assert is_plugin_in_deactivation_list if session_state == "child" else not is_plugin_in_deactivation_list
    elif plugin_parallel_mode == parallel_utils.ParallelPluginModes.CHILD_ONLY:
        assert is_plugin_in_deactivation_list if session_state == "parent" else not is_plugin_in_deactivation_list


@pytest.mark.parametrize('is_parallel_mode', [True, False])
def test_parallel_mode_activation_causes_warning(config_override, checkpoint, is_parallel_mode):

    @slash.plugins.parallel_mode(parallel_utils.ParallelPluginModes.DISABLED)
    class Plugin(slash.plugins.PluginInterface):
        def session_start(self):
            checkpoint()

        def get_name(self):
            return type(self).__name__.lower()

    if is_parallel_mode:
        config_override("parallel.num_workers", 1)
        assert parallel_utils.is_parallel_session()
    plugin = Plugin()
    assert try_get_mark(plugin, 'parallel_mode') == parallel_utils.ParallelPluginModes.DISABLED
    with Session():
        slash.plugins.manager.install(plugin)
        slash.plugins.manager.activate(plugin)
        slash.hooks.session_start() # pylint: disable=no-member

        warning_message_expected = "Activating plugin {} though it's configuration for parallel mode doesn't fit to current session"\
                                    .format(plugin.get_name())
        is_warning_appeared = warning_message_expected in [warning.message for warning in slash.session.warnings]
        assert is_warning_appeared if is_parallel_mode else not is_warning_appeared
        assert checkpoint.called



#!/usr/bin/env python
import argparse
import logging
import os
import sys
import subprocess
import time

def _from_root(*p):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *p))

_VERSION_FILE = _from_root("slash", "__version__.py")
_CHANGELOG_FILE = _from_root("doc", "changelog.rst")

parser = argparse.ArgumentParser(usage="%(prog)s [options] args...")
parser.add_argument("-v", action="append_const", const=1, dest="verbosity", default=[],
                    help="Be more verbose. Can be specified multiple times to increase verbosity further")


def main(args):
    default = _get_default_next_version()
    version = raw_input("New version to release (default: {!r}): ".format(default)).strip()
    if not version:
        version = default
    subprocess.check_call('git flow release start {}'.format(version), shell=True)
    _write_new_version(version)
    _write_changelog(version)
    subprocess.check_call("git commit -a -m 'v{}'".format(version), shell=True)
    subprocess.check_call('git flow release finish {}'.format(version), shell=True)
    return 0

def _get_default_next_version():
    major, minor, bugfix = _get_current_version()
    return "{}.{}.{}".format(major, minor, bugfix + 1)

def _get_current_version():
    d = {}
    with open(_VERSION_FILE) as version_file:
        exec(version_file.read(), d, d)

    return [int(x) for x in d["__version__"].split(".")]

def _write_new_version(version):
    with open(_VERSION_FILE, "w") as version_file:
        version_file.write("__version__ = {!r}\n".format(version))

def _write_changelog(version):
    temp_filename = "/tmp/__bump_version_changelog.rst"
    with open(_CHANGELOG_FILE) as changelog_file:
        with open(temp_filename, "w") as temp_file:
            already_wrote = False
            for line in changelog_file:
                if line.startswith("* ") and not already_wrote:
                    temp_file.write("* :release:`{} <{}>`\n".format(
                        version,
                        time.strftime("%d-%m-%Y"),
                        ))
                    already_wrote = True
                temp_file.write(line)
    os.rename(temp_filename, _CHANGELOG_FILE)

#### For use with entry_points/console_scripts
def main_entry_point():
    args = parser.parse_args()
    sys.exit(main(args))


if __name__ == "__main__":
    main_entry_point()



import pytest
import slash.plugins
from .conftest import Checkpoint
from .utils import maybe_decorate
from slash.plugins import PluginInterface
from gossip.exceptions import CannotResolveDependencies


@pytest.mark.parametrize('needs_decorate_method', [True, False])
@pytest.mark.parametrize('provides_decorate_method', [True, False])
def test_needs_provides_plugin_name(needs_decorate_method, provides_decorate_method, checkpoint1, checkpoint2):

    @slash.plugins.active  # pylint: disable=abstract-method, unused-variable
    @maybe_decorate(slash.plugins.needs('p'), not needs_decorate_method)
    @autoname
    class NeedsPlugin(PluginInterface):

        @maybe_decorate(slash.plugins.needs('p'), needs_decorate_method)
        def session_start(self):
            checkpoint2()

    @slash.plugins.active  # pylint: disable=abstract-method, unused-variable
    @maybe_decorate(slash.plugins.provides('p'), not provides_decorate_method)
    @autoname
    class ProvidesPlugin(PluginInterface):

        @maybe_decorate(slash.plugins.provides('p'), provides_decorate_method)
        def session_start(self):
            checkpoint1()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp


def test_provides_globally_needs_globally(checkpoint1, checkpoint2):
    '''
    Plugin A: Provides x at class level
    Plugin B: Needs x at class level
    '''
    @slash.plugins.provides('x')
    class PluginA(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_a'

        def session_start(self):
            checkpoint1()

        def test_start(self):
            pass


    @slash.plugins.needs('x')
    class PluginB(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_b'

        def session_start(self):
            checkpoint2()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


def test_provides_globally_needs_specific_hook(checkpoint1, checkpoint2):
    '''
    Plugin A: Provides x at class level
    Plugin B: Needs x for specific hook
    '''
    @slash.plugins.provides('x')
    class PluginA(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_a'

        def session_start(self):
            checkpoint1()

        def test_start(self):
            pass


    class PluginB(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_b'

        @slash.plugins.needs('x')
        def session_start(self):
            checkpoint2()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


def test_provides_globally_needs_specific_hook_which_does_not_exist_at_a(checkpoint2):
    '''
    Plugin A: Provides x at class level
    Plugin B: Needs x for specific hook, this hook does not definied in A

    Expectations:
    Should work in the empty sense
    all non-needing hooks should work, even when missing from A, the specific hook needs to happen in A before B.
    '''
    @slash.plugins.provides('x')
    class PluginA(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_a'

        def test_start(self):
            pass


    class PluginB(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_b'

        @slash.plugins.needs('x')
        def session_start(self):
            checkpoint2()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint2.called

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


def test_provides_specific_hook_needs_globally(checkpoint1, checkpoint2):
    '''
    Plugin A: Provides x on a specific hook
    Plugin B: Needs x at class level

    Expectations:
    This case should fail, because logically the other hooks don't have anyone to provide X for them
    '''
    class PluginA(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_a'

        @slash.plugins.provides('x')
        def session_start(self):
            checkpoint1()

        def test_start(self):
            pass


    @slash.plugins.needs('x')
    class PluginB(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_b'

        def session_start(self):
            checkpoint2()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


def test_provides_specific_hook_needs_globally_with_this_hook_only(checkpoint1, checkpoint2):
    '''
    Plugin A: Provides x on a specific hook
    Plugin B: Needs x at class level, but only has one hook (the one provided by A)
    '''
    class PluginA(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_a'

        @slash.plugins.provides('x')
        def session_start(self):
            checkpoint1()

        def test_start(self):
            pass


    @slash.plugins.needs('x')
    class PluginB(slash.plugins.interface.PluginInterface):

        def get_name(self):
            return 'plugin_b'

        def session_start(self):
            checkpoint2()

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


@pytest.mark.parametrize('needs_parent_level', [True, False])
@pytest.mark.parametrize('provides_parent_level', [True, False])
def test_provides_needs_with_inheritence_on_class_level(checkpoint, checkpoint1, checkpoint2, needs_parent_level, provides_parent_level):
    '''
    Plugin A: Provides x in class level (by it self or by inheritence)
    Plugin b: Needs x in class level (by it self or by inheritence)
    '''
    # pylint: disable=abstract-method

    @maybe_decorate(slash.plugins.provides('x'), provides_parent_level)
    class PluginAParent(slash.plugins.interface.PluginInterface):

        def test_start(self):
            pass

    @maybe_decorate(slash.plugins.provides('x'), not provides_parent_level)
    class PluginA(PluginAParent):

        def get_name(self):
            return 'plugin_a'

        def session_start(self):
            checkpoint1()


    @maybe_decorate(slash.plugins.needs('x'), needs_parent_level)
    class PluginBParent(slash.plugins.interface.PluginInterface):

        def error_added(self, result, error): # pylint: disable=unused-argument
            checkpoint()


    @maybe_decorate(slash.plugins.needs('x'), not needs_parent_level)
    class PluginB(PluginBParent):

        def get_name(self):
            return 'plugin_b'

        def session_start(self):
            checkpoint2()

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    # session_start hook should be provided the PluginA.session_start method
    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp

    # error_added hook should be provided by empty registration of pluginA
    slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert checkpoint.called

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.error_added()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])

    # Ensure only hooks required by PluginB fails
    slash.hooks.test_end()  # pylint: disable=no-member


def test_provides_needs_in_both_inheritence_levels(checkpoint, checkpoint1, checkpoint2):
    # pylint: disable=abstract-method

    @slash.plugins.provides('x')
    class PluginAParent(slash.plugins.interface.PluginInterface):

        def test_start(self):
            pass

    @slash.plugins.provides('y')
    class PluginA(PluginAParent):

        def get_name(self):
            return 'plugin_a'

        def session_start(self):
            checkpoint1()


    @slash.plugins.needs('x')
    class PluginBParent(slash.plugins.interface.PluginInterface):

        def error_added(self, result, error): # pylint: disable=unused-argument
            checkpoint()


    @slash.plugins.needs('y')
    class PluginB(PluginBParent):

        def get_name(self):
            return 'plugin_b'

        def session_start(self):
            checkpoint2()

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    # session_start hook should be provided the PluginA.session_start method
    slash.hooks.session_start()  # pylint: disable=no-member
    assert checkpoint1.timestamp < checkpoint2.timestamp

    # error_added hook should be provided by empty registration of pluginA
    slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert checkpoint.called

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.error_added()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    # Ensure only hooks required by PluginB fails
    slash.hooks.test_end()  # pylint: disable=no-member


def test_provides_needs_with_inheritence_on_method_level(checkpoint):
    '''
    Plugin A: Provides x in method level (by it self or by inheritence) to test_start & session_start
    Plugin b: Needs x in method level (by it self or by inheritence) on test_start & session_start
    '''
    # pylint: disable=abstract-method
    session_start_a = Checkpoint()
    session_start_b = Checkpoint()
    test_start_a = Checkpoint()
    test_start_b = Checkpoint()

    class PluginAParent(slash.plugins.interface.PluginInterface):

        @slash.plugins.provides('x')
        def test_start(self):
            test_start_a()

    class PluginA(PluginAParent):

        def get_name(self):
            return 'plugin_a'

        @slash.plugins.provides('x')
        def session_start(self):
            session_start_a()


    class PluginBParent(slash.plugins.interface.PluginInterface):

        @slash.plugins.needs('x')
        def session_start(self):
            session_start_b()

        def error_added(self, result, error): # pylint: disable=unused-argument
            checkpoint()


    class PluginB(PluginBParent):

        def get_name(self):
            return 'plugin_b'

        @slash.plugins.needs('x')
        def test_start(self):
            test_start_b()

    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert session_start_a.timestamp < session_start_b.timestamp

    slash.hooks.test_start()  # pylint: disable=no-member
    assert test_start_a.timestamp < test_start_b.timestamp

    # error_added hook should not need anything
    slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert checkpoint.called

    slash.plugins.manager.deactivate('plugin_a')
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])
    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.test_start()  # pylint: disable=no-member
    slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x'])


def test_provides_needs_with_child_overrides():
    # pylint: disable=line-too-long
    '''
    | Hook Name     | Plugin A                                                                | Plugin B                                                               |
    |---------------+-------------------------------------------------------------------------+------------------------------------------------------------------------|
    | session_start | Child Provides x in method level, overrides parent's empty registration | Needs x (Parent) & y (Child) in class level                            |
    | test_start    | Child Provides x in method level, overrides parent's real registration  | Needs x (Parent) & y (Child) in class level                            |
    | error_added   | x is not provided, overrides parent's real registration                 | Needs x (Parent) & y (Child) in class level                            |
    | test_end      | x is not provided, overrides parent's empty registration                | Needs x (Parent) & y (Child) in class level                            |
    | session_end   | Parent provides x, child provides y - both in class level               | Needs x (Parent) & y (Child) in class level, z in (child) method level |
    '''
    # pylint: disable=abstract-method
    session_start_a = Checkpoint()
    session_start_b = Checkpoint()
    test_start_a = Checkpoint()
    test_start_b = Checkpoint()

    @slash.plugins.provides('x')
    class PluginAParent(slash.plugins.interface.PluginInterface):

        def test_start(self):
            test_start_a()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

        def session_end(self):
            pass

    @slash.plugins.provides('y')
    class PluginA(PluginAParent):

        def get_name(self):
            return 'plugin_a'

        @slash.plugins.provides('x')
        def session_start(self):
            # Overrides empty registration of PluginAParent
            session_start_a()

        @slash.plugins.provides('x')
        def test_start(self):
            # Overrides "real" registration of PluginAParent
            test_start_a()

        def error_added(self, result, error): # pylint: disable=unused-argument
            # Overrides "real" registration of PluginAParent
            pass

        def test_end(self):
            # Overrides empty registration of PluginAParent
            pass


    @slash.plugins.needs('x')
    class PluginBParent(slash.plugins.interface.PluginInterface):

        def session_start(self):
            session_start_b()

        def error_added(self, result, error): # pylint: disable=unused-argument
            pass

        def test_start(self):
            test_start_b()

        def test_end(self):
            pass


    @slash.plugins.needs('y')
    class PluginB(PluginBParent):

        def get_name(self):
            return 'plugin_b'

        @slash.plugins.needs('z')
        def session_end(self):
            pass


    for plugin_cls in [PluginA, PluginB]:
        slash.plugins.manager.install(plugin_cls(), activate_later=True)
    slash.plugins.manager.activate_pending_plugins()

    slash.hooks.session_start()  # pylint: disable=no-member
    assert session_start_a.timestamp < session_start_b.timestamp

    slash.hooks.test_start()  # pylint: disable=no-member
    assert test_start_a.timestamp < test_start_b.timestamp

    slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member

    slash.hooks.test_end()  # pylint: disable=no-member

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_end()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['z'])


    slash.plugins.manager.deactivate('plugin_a')

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.test_start()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.error_added(result=None, error=None)  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.test_end()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y'])

    with pytest.raises(CannotResolveDependencies) as caught:
        slash.hooks.session_end()  # pylint: disable=no-member
    assert caught.value.unmet_dependencies == set(['x', 'y', 'z'])


def autoname(plugin):
    def get_name(self):
        return type(self).__name__.lower()
    plugin.get_name = get_name
    return plugin



import itertools
import re

import logbook

from slash._compat import iteritems, itervalues, xrange

from .generator_fixture import GeneratorFixture

_logger = logbook.Logger(__name__)


def validate_run(suite, run_result, expect_interruption, expect_session_errors):
    if expect_interruption or not run_result.session.results.is_success(allow_skips=True):
        assert run_result.exit_code != 0, '`slash run` unexpectedly returned 0'
    else:
        assert run_result.exit_code == 0, '`slash run` unexpectedly returned {}. Output: {}'.format(
            run_result.exit_code, run_result.get_console_output())

    global_result = run_result.session.results.global_result
    errors = global_result.get_errors() + global_result.get_failures()
    if expect_session_errors:
        assert errors, 'Expected session errors but found none'
    else:
        assert not errors, 'Sessions errors were not expected (Got {})'.format(errors)

    for test, results in iteritems(_group_results_by_test_id(suite, run_result)):
        _validate_single_test(test, results)


def _validate_single_test(test, results):

    param_names = {p.id: p.name for p in _find_all_parameters(test)}

    for param_values in _iter_param_value_sets(test):

        is_excluded = any((param_names[param_id], value) in test.excluded_param_values for param_id, value in param_values.items())

        for repetition in xrange(test.get_num_expected_repetitions()):  # pylint: disable=unused-variable

            for index, result in enumerate(results):

                if _result_matches(result, param_values):

                    results.pop(index)

                    if is_excluded:
                        assert result.is_skip()
                    else:
                        _validate_single_test_result(test, result)

                    break
            else:
                assert False, 'Could not find parameter set {}'.format(
                    param_values)

    assert not results, 'Unmatched results exist'


def _iter_param_value_sets(test):
    params = _find_all_parameters(test)
    param_ids = [p.id for p in params]
    for combination in itertools.product(*(param.values for param in params)):
        yield dict(zip(param_ids, combination))


def _find_all_parameters(func):
    params = []
    stack = [func]
    while stack:
        f = stack.pop()
        for subfixture in f.get_fixtures():
            if isinstance(subfixture, GeneratorFixture):
                params.append(subfixture)
                continue
            else:
                stack.append(subfixture)
        params.extend(f.get_parameters())
    return list(itervalues({p.id: p for p in params}))


def _result_matches(result, param_values):
    values = result.test_metadata.variation.values.copy()
    for param_name in list(values):
        # handle the case of a fixture with a single param, which is logically a parameter by itself
        if re.match(r'^fx_\d+.param$', param_name):
            values_name = param_name.split('_')[1].split('.')[0]
        else:
            values_name = param_name.rsplit('_', 1)[-1]

        values[values_name] = values.pop(param_name)

    return values == param_values


def _validate_single_test_result(test, result):
    expected = test.get_expected_result()
    if expected == 'ERROR':
        assert result.is_error(), 'Test did not issue error as expected'
    elif expected == 'FAIL':
        assert result.is_failure(), 'Test did not fail as expected'
    elif expected == 'SUCCESS':
        assert result.is_success(), 'Test {} unexpectedly unsuccessful:\n{}'.format(
            test.id, list(itertools.chain(result.get_errors(), result.get_failures())))
    elif expected == 'INTERRUPT':
        assert result.is_interrupted(), 'Test did not get interrupted as expected'
    elif expected == 'SKIP':
        assert result.is_skip()
    elif expected == 'NOT_RUN':
        assert result.is_not_run()
    else:
        raise NotImplementedError(
            'Unknown expected result: {!r}'.format(expected))  # pragma: no cover


def _group_results_by_test_id(suite, run_result):
    tests_by_id = dict((t.id, t) for t in suite)
    unseen = tests_by_id.copy()

    groups = {}

    for result in run_result.session.results:
        if result.test_metadata.is_interactive():
            continue
        test_id = get_test_id_from_test_address(result.test_metadata.address)
        assert tests_by_id[test_id].is_selected(), 'Test {} appears in results, although not expected!'.format(test_id)
        groups.setdefault(tests_by_id[test_id], []).append(result)
        unseen.pop(test_id, None)

    for test_id, test in list(iteritems(unseen)):
        if not test.is_selected():
            unseen.pop(test_id, None)

    assert not unseen, 'Expected results not found ({})'.format(unseen)

    return groups


def get_test_id_from_test_address(addr):
    _, addr = addr.split(':', 1)
    return addr.split('_')[1].split('(')[0]



from __future__ import print_function

import argparse
import inspect
import itertools
import os
import sys
from functools import partial

import colorama
import slash
import slash.site
import slash.loader
from slash import config
from slash.exceptions import CannotLoadTests
from slash.utils.cli_utils import UNDERLINED, make_styler, Printer
from slash.utils.python import get_underlying_func
from slash.utils.suite_files import iter_suite_file_paths

_heading_style = make_styler(colorama.Fore.MAGENTA + colorama.Style.BRIGHT + UNDERLINED)  # pylint: disable=no-member
_title_style = make_styler(colorama.Fore.WHITE + colorama.Style.BRIGHT)  # pylint: disable=no-member
_unused_style = make_styler(colorama.Fore.YELLOW)  # pylint: disable=no-member
_doc_style = make_styler(colorama.Fore.GREEN + colorama.Style.BRIGHT)  # pylint: disable=no-member
_override_style = make_styler(colorama.Fore.YELLOW + colorama.Style.BRIGHT)  # pylint: disable=no-member


def _get_parser():
    parser = argparse.ArgumentParser('slash list [options] PATH...')
    parser.add_argument('-f', '--suite-file', dest='suite_files', action='append', default=[])
    parser.add_argument('--only-fixtures', dest='only', action='store_const', const='fixtures', default=None)
    parser.add_argument('--only-tests', dest='only', action='store_const', const='tests', default=None)
    parser.add_argument('--show-tags', dest='show_tags', action='store_true', default=False)
    parser.add_argument('--no-params', dest='show_params', action='store_false', default=True)
    parser.add_argument('--allow-empty', dest='allow_empty', action='store_true', default=False)
    parser.add_argument('--warnings-as-errors', dest='warnings_as_errors', action='store_true', default=False)
    parser.add_argument('-r', '--relative-paths', action='store_true', default=False)
    parser.add_argument('--no-output', dest='show_output', action='store_false', default=True)
    parser.add_argument('--force-color', dest='force_color', action='store_true', default=False)
    parser.add_argument('--no-color', dest='enable_color', action='store_false', default=True)

    parser.add_argument('paths', nargs='*', default=[], metavar='PATH')
    return parser


def slash_list(args, report_stream=sys.stdout, error_stream=sys.stderr):
    _print = partial(print, file=report_stream)

    parser = _get_parser()
    parsed_args = parser.parse_args(args)

    _print = Printer(report_stream, enable_output=parsed_args.show_output, force_color=parsed_args.force_color,
                     enable_color=parsed_args.enable_color)
    try:
        with slash.Session() as session:
            slash.site.load()

            if not parsed_args.paths and not parsed_args.suite_files:
                parsed_args.paths = config.root.run.default_sources

            if not parsed_args.paths and not parsed_args.suite_files:
                parser.error('Neither test paths nor suite files were specified')

            loader = slash.loader.Loader()
            runnables = loader.get_runnables(itertools.chain(parsed_args.paths, iter_suite_file_paths(parsed_args.suite_files)))
            used_fixtures = set()
            for test in runnables:
                used_fixtures.update(test.get_required_fixture_objects())

            if parsed_args.only in (None, 'fixtures'):
                _report_fixtures(parsed_args, session, _print, used_fixtures)

            if parsed_args.only in (None, 'tests'):
                _report_tests(parsed_args, runnables, _print)

        if bool(session.warnings.warnings) and parsed_args.warnings_as_errors:
            return -1
        if len(runnables):  # pylint: disable=len-as-condition
            return 0
    except CannotLoadTests as e:
        print('Could not load tests ({})'.format(e), file=error_stream)
        return -1
    print('No tests were found!', file=sys.stderr)
    return int(not parsed_args.allow_empty)


def _report_tests(args, runnables, printer):
    if not args.only:
        printer(_heading_style('Tests'))

    visited = set()

    for runnable in runnables:
        extra = "" if not args.show_tags else "  Tags: {}".format(list(runnable.get_tags()))
        address = runnable.__slash__.address
        if not args.show_params:
            address = address.split('(')[0]
        if args.relative_paths:
            address = _convert_address_to_relpath(address)
        if address in visited:
            continue
        visited.add(address)
        printer("{}{}".format(_title_style(address), extra))


def _convert_address_to_relpath(address):
    filename, remainder = address.split(':', 1)
    if os.path.isabs(filename):
        filename = os.path.relpath(filename)
    return '{}:{}'.format(filename, remainder)


def _report_fixtures(args, session, printer, used_fixtures):
    if not args.only:
        printer(_heading_style('Fixtures'))
    for fixture in session.fixture_store:
        if not hasattr(fixture, 'fixture_func'):
            continue

        fixture_func = get_underlying_func(fixture.fixture_func)
        doc = inspect.cleandoc(fixture_func.__doc__) if fixture_func.__doc__ else ''

        if fixture.info.autouse:
            additional_info = ' (Autouse)'
        elif fixture not in used_fixtures:
            additional_info = ' (Unused)'
        else:
            additional_info = ''

        if fixture.is_override():
            additional_info += ' -- ' + _override_style('Override')

        printer(_title_style('{}{}'.format(fixture.info.name, additional_info)))
        if doc:
            for line in (_doc_style(doc)).split('\n'):
                printer('    {}'.format(line))

        printer('    Source: {}:{}'.format(
            os.path.relpath(inspect.getsourcefile(fixture_func), args.paths[0] if args.paths else '.'),
            inspect.getsourcelines(fixture_func)[1]))
        printer('\n')



from slash.frontend.list_config import list_config
from slash._compat import StringIO


def test_slash_list_config():
    report_stream = StringIO()
    list_config([], report_stream)
    assert report_stream.getvalue()

def test_slash_list_config_with_filters():
    report_stream = StringIO()
    list_config(['log'], report_stream)
    assert 'log.root' in report_stream.getvalue()



import os
from uuid import uuid4

from slash._compat import izip_longest
from slash.frontend.slash_run import slash_run

from .suite_writer.utils import get_code_lines

class SuiteBuilder(object):

    def __init__(self, path):
        super(SuiteBuilder, self).__init__()
        self.path = path
        self.files = [SuiteBuilderFile()]

    def build(self):
        returned = SuiteBuilderSuite(os.path.join(self.path, 'suite-{}'.format(uuid4())))
        for index, file in enumerate(self.files):
            with open(os.path.join(returned.path, 'test_file_{}.py'.format(index)), 'w') as f:
                for code in file.code_snippets:
                    for line in get_code_lines(code):
                        f.write(line)
                        f.write('\n')
        return returned

    @property
    def first_file(self):
        return self.files[0]


class SuiteBuilderFile(object):

    def __init__(self):
        self.code_snippets = []

    def add_code(self, code):
        self.code_snippets.append(code)
        return self


class SuiteBuilderSuite(object):

    def __init__(self, path):
        self.path = path
        os.makedirs(path)

    def run(self):
        app = slash_run([self.path])
        return SuiteBuilderSuiteResult(app)


class SuiteBuilderSuiteResult(object):

    def __init__(self, slash_app):
        self.slash_app = slash_app

    def assert_success(self, num_tests):
        return self.assert_all(num_tests).success()

    def assert_all(self, num_tests):
        assert len(self.slash_app.session.results) == num_tests
        return AssertAllHelper(self)

    def assert_results(self, num_results):
        results = list(self.slash_app.session.results.iter_test_results())
        assert len(results) == num_results
        return results

    def assert_session_error(self, error_substring):
        self.assert_all(0)
        errs = self.slash_app.session.results.global_result.get_errors()
        assert len(errs) == 1
        assert error_substring in str(errs[0])

    def with_data(self, data_sets):
        results = list(self.slash_app.session.results)
        data_sets = list(data_sets)
        while data_sets:
            data_set = data_sets.pop()
            for index, result in enumerate(results):
                if result.data == data_set:
                    results.pop(index)
                    break
            else:
                assert False, 'No result found for {}'.format(data_set)
        assert not results
        return self

class AssertAllHelper(object):

    def __init__(self, suite_builder_result):
        self.suite_builder_result = suite_builder_result
        self._results = suite_builder_result.slash_app.session.results

    def success(self):
        assert self._results.is_success(allow_skips=False)
        return self.suite_builder_result

    def errors(self, errors_list):
        for res, error in izip_longest(self._results, errors_list):
            errs = res.get_errors()
            assert len(errs) == 1
            assert errs[0] == error
        return self.suite_builder_result



# pylint: disable=redefined-outer-name, unused-argument
from __future__ import print_function

import functools
import os
import random
from uuid import uuid4

import pytest

import slash


def test_normal_sorting(test_dir, names):
    assert get_file_names(load(test_dir)) == names

def test_custom_ordering(test_dir, names, indices):
    @slash.hooks.tests_loaded.register # pylint: disable=no-member
    def tests_loaded(tests):                # pylint: disable=unused-variable
        for index, (test, new_index) in enumerate(zip(tests, indices)):
            assert test.__slash__.file_path.endswith(names[index]), 'Tests are loaded in incorrect order'
            test.__slash__.set_sort_key(new_index)
    assert get_file_names(load(test_dir)) == _sorted_by_indices(names, indices)

################################################################################
## Utils and fixtures

def get_file_names(tests):
    returned = []
    for t in tests:
        file_path = t.__slash__.file_path
        assert os.path.isabs(file_path)
        returned.append(os.path.join(
            os.path.basename(os.path.dirname(file_path)),
            os.path.basename(file_path)))
    return returned

def load(source):
    with slash.Session():
        return slash.loader.Loader().get_runnables([source])

@pytest.fixture
def test_dir(tmpdir, names, indices):
    returned = tmpdir.join(str(uuid4()))
    indices = range(len(names))[::-1]
    for index, name in zip(indices, names):
        with returned.join(name).open('w', ensure=True) as f:
            _print = functools.partial(print, file=f)
            _print('import slash')
            _print('@slash.tag("index", {})'.format(index))
            _print('def test_something():')
            _print('    pass')
    return str(returned)

@pytest.fixture
def names():
    return ['a/test_a_b.py', 'a/test_b_a.py', 'b/test_a_a.py', 'b/test_a_b.py']

def _sorted_by_indices(items, indices):
    returned = [None for _ in items]
    for index, item in zip(indices, items):
        returned[index] = item
    return returned

def _randomized(l):
    indices = list(range(len(l)))
    random.shuffle(indices)
    return [l[index] for index in indices]

@pytest.fixture(params=[reversed, _randomized])
def indices(request, names):
    return [names.index(name) for name in request.param(names)]



from .utils import (
    CustomException,
)
from contextlib import contextmanager

import pytest

import slash
from slash import should
from slash.exceptions import TestFailed, ExpectedExceptionNotCaught


@pytest.mark.parametrize('pair', [
    (1, 1),
    (1, 1.00000001),
])
def test_assert_almost_equal_positive(pair):
    a, b = pair
    slash.assert_almost_equal(a, b)


@pytest.mark.parametrize('combination', [
    (1, 1, 0),
    (1, 1, 0.1),
    (1, 1.1, 0.5),
    (1.0001, 1.00009, 0.00002),
])
def test_assert_almost_equal_positive_with_delta(combination):
    a, b, delta = combination
    slash.assert_almost_equal(a, b, delta)


@pytest.mark.parametrize('combination', [
    (1, 12, 0),
    (1, 1.1, 0.00001),
    (1.0001, 1.00009, 0.000001),
])
def test_assert_almost_equal_negative_with_delta(combination):
    a, b, delta = combination
    with pytest.raises(AssertionError):
        slash.assert_almost_equal(a, b, delta)


def test_assert_equals():
    with checking(should.equal, should.not_equal):
        good(1, 1)
        bad(1, 2)


def test_assert_empty():
    with checking(should.be_empty, should.not_be_empty):
        good([])
        good({})
        good(set())
        bad([1, 2, 3])
        bad({1: 2})
        bad(set([1, 2, 3]))


def test_assert_isinstance():
    with checking(should.be_a, should.not_be_a):
        good(1, int)
        good("a", str)
        good({}, dict)
        bad(1, 1)
        bad(1, str)
        bad(None, str)


def test_is_none():
    with checking(should.be_none, should.not_be_none):
        good(None)
        bad("None")


def test_is():
    obj = object()
    with checking(should.be, should.not_be):
        good(obj, obj)
        bad(obj, object())
        bad({}, {})


def test_truth():
    with checking(should.be_true, should.be_false):
        good(True)
        good("hello")
        bad(False)
        bad("")
        bad({})


def test_in():
    with checking(should.be_in, should.not_be_in):
        good(1, [1, 2, 3])
        good("e", "hello")
        bad(1, [])
        bad("e", "fdfd")
    with checking(should.contain, should.not_contain):
        good([1, 2, 3], 1)
        good("hello", "e")
        bad("fdfdfd", "e")



@pytest.mark.parametrize('func', [should.raise_exception, slash.assert_raises])
def test_assert_raises(func):
    thrown = CustomException()
    with func(CustomException) as caught:
        raise thrown
    assert caught.exception is thrown
    try:
        with func(CustomException):
            raise OtherException()
    except OtherException:
        pass
    else:
        assert False, "func allowed a different type of exception to be raised"
    try:
        with func(CustomException):
            pass
    except ExpectedExceptionNotCaught as e:
        assert " not raised" in str(e)
        assert e.expected_types == (CustomException,)
    else:
        assert False, "should.raise_exception allowed success"


@pytest.mark.parametrize('func', [should.raise_exception, slash.assert_raises])
def test_assert_raises_multiple_exceptions(func):
    class CustomException1(Exception):
        pass

    class CustomException2(Exception):
        pass

    class CustomException3(Exception):
        pass

    exception_types = (CustomException1, CustomException2)

    for exc_type in exception_types:
        with func(exception_types) as caught:
            value = exc_type('!')
            raise value
        assert caught.exception is value

    value = CustomException3('!')
    with pytest.raises(CustomException3) as caught:
        with func(exception_types):
            raise value
    assert caught.value is value




def test_raises_exception_multiple_classes():
    possible_exception_types = (CustomException, OtherException)
    for x in possible_exception_types:
        with should.raise_exception(possible_exception_types):
            raise x()

    with pytest.raises(ExpectedExceptionNotCaught):
        with should.raise_exception((CustomException,)):
            pass

# boilerplate


class OtherException(BaseException):
    pass

_MESSAGE = "SOME MESSAGE HERE"


def good(*args):
    _test_assertion(args, positive=True)


def bad(*args):
    _test_assertion(args, positive=False)


def _test_assertion(args, positive):
    positive_assertion = _current_positive_assertion
    negative_assertion = _current_negative_assertion
    if not positive:
        positive_assertion, negative_assertion = negative_assertion, positive_assertion
    positive_assertion(*args)
    positive_assertion(*(args + (_MESSAGE,)))
    for msg in (None, _MESSAGE):
        try:
            negative_assertion(*(args + (msg,)))
        except TestFailed as e:
            if msg is not None:
                assert msg in str(e)
        else:
            assert False, "Assertion did not fail"


_current_positive_assertion = None
_current_negative_assertion = None


@contextmanager
def checking(assertion, negative_assertion):
    global _current_positive_assertion # pylint: disable=global-statement
    global _current_negative_assertion # pylint: disable=global-statement
    assert _current_positive_assertion is None
    _current_positive_assertion = assertion
    _current_negative_assertion = negative_assertion
    try:
        yield
    finally:
        _current_positive_assertion = None
        _current_negative_assertion = None



# pylint: disable=redefined-outer-name,unused-argument
import sys
import pytest
import coverage

import slash.plugins


@pytest.mark.skipif(hasattr(sys, 'pypy_version_info'), reason='Cannot run on pypy')
def test_coverage_plugin(suite, enabled_coverage_plugin, config_override, report_type):
    config_override('plugin_config.coverage.report', True)
    config_override('plugin_config.coverage.report_type', report_type)
    suite.run()

def test_invalid_report_type(config_override):
    config_override('plugin_config.coverage.report_type', 'bla')
    with pytest.raises(RuntimeError) as exc:
        slash.plugins.manager.activate('coverage')
    assert 'Unknown report type' in str(exc.value)


@pytest.fixture
def enabled_coverage_plugin(request, patched_coverage, config_override, should_append, report_type):
    config_override('plugin_config.coverage.report', False)
    config_override('plugin_config.coverage.append', should_append)
    config_override('plugin_config.coverage.report_type', report_type)
    slash.plugins.manager.activate('coverage')

    @request.addfinalizer
    def deactivate():  # pylint: disable=unused-variable
        slash.plugins.manager.deactivate('coverage')


@pytest.fixture(params=[True, False])
def should_append(request):
    return request.param


@pytest.fixture(params=['html', 'xml', 'html,xml'])
def report_type(request):
    return request.param


@pytest.yield_fixture
def patched_coverage(forge):
    forge.replace_with(coverage.Coverage, 'start', lambda self: None)
    forge.replay()
    try:
        yield
    finally:
        forge.restore_all_replacements()
